import argparse
import enum
import functools
import io
import itertools
import logging
import os
import shutil
import subprocess
import sys
import time
import urllib.parse
import urllib.request
from collections import defaultdict
from pathlib import Path
from typing import Optional, Union

from dissect.target import Target, exceptions
from dissect.target.filesystems import dir, ntfs
from dissect.target.helpers import fsutil
from dissect.target.loaders.remote import RemoteStreamConnection
from dissect.target.plugins.apps.webservers import iis
from dissect.target.plugins.os.windows.log import evt, evtx

from acquire.collector import Collector, get_full_formatted_report, get_report_summary
from acquire.dynamic.windows.named_objects import NamedObjectType
from acquire.esxi import esxi_memory_context_manager
from acquire.hashes import (
    HashFunc,
    collect_hashes,
    filter_out_by_path_match,
    filter_out_by_value_match,
    filter_out_huge_files,
    serialize_into_csv,
)
from acquire.log import get_file_handler, reconfigure_log_file, setup_logging
from acquire.outputs import OUTPUTS
from acquire.uploaders.minio import MinIO
from acquire.uploaders.plugin import UploaderPlugin, upload_files_using_uploader
from acquire.uploaders.plugin_registry import UploaderRegistry
from acquire.utils import (
    check_and_set_acquire_args,
    check_and_set_log_args,
    create_argument_parser,
    format_output_name,
    get_formatted_exception,
    get_user_name,
    get_utc_now,
    get_utc_now_str,
    is_user_admin,
    normalize_path,
    parse_acquire_args,
    persist_execution_report,
)

try:
    from acquire.version import version
except ImportError:
    version = "0.0.dev"

try:
    # Injected by pystandalone builder
    from acquire.config import CONFIG
except ImportError:
    CONFIG = defaultdict(lambda: None)


VERSION = version
ACQUIRE_BANNER = r"""
                       _
  __ _  ___ __ _ _   _(_)_ __ ___
 / _` |/ __/ _` | | | | | '__/ _ \
| (_| | (_| (_| | |_| | | | |  __/
 \__,_|\___\__, |\__,_|_|_|  \___|
  by Fox-IT   |_|             v{}
  part of NCC Group
""".format(
    VERSION
)[
    1:
]

MODULES = {}
MODULE_LOOKUP = {}

CLI_ARGS_MODULE = "cli-args"


log = logging.getLogger("acquire")
log.propagate = 0
log_file_handler = None
logging.lastResort = None
logging.raiseExceptions = False


def misc_windows_user_homes(target):
    misc_dirs = {
        ("windows/serviceprofiles/localservice", False),
        ("windows/serviceprofiles/networkservice", False),
        ("windows/system32/config/systemprofile", False),
        ("users", True),
        ("documents and settings", True),
    }

    for fs in target.fs.path().iterdir():
        if fs.name.lower() == "c:":
            continue

        for misc_dir, get_subdirs in misc_dirs:
            misc_path = fs.joinpath(misc_dir)

            if not misc_path.exists():
                continue

            if get_subdirs:
                for entry in misc_path.iterdir():
                    if entry.is_dir():
                        yield entry
            else:
                yield misc_path


def from_user_home(target, path):
    for user_details in target.user_details.all_with_home():
        yield str(user_details.home_path.joinpath(path))
    if target.os == "windows":
        for misc_dir in misc_windows_user_homes(target):
            yield str(misc_dir.joinpath(path))


def iter_ntfs_filesystems(target):
    mount_lookup = defaultdict(list)
    for mount, fs in target.fs.mounts.items():
        mount_lookup[fs].append(mount)

    sysvol = target.fs.mounts["sysvol"]
    for fs in target.filesystems:
        if fs in mount_lookup:
            mountpoints = ", ".join(mount_lookup[fs])
        else:
            mountpoints = "No mounts"

        if not isinstance(fs, ntfs.NtfsFilesystem):
            log.warning("Skipping %s (%s) - not an NTFS filesystem", fs, mountpoints)
            continue

        if fs == sysvol:
            name = "sysvol"
        elif fs in mount_lookup:
            name = mount_lookup[fs][0]
        else:
            name = f"vol-{fs.ntfs.serial:x}"

        yield fs, name, mountpoints


def mount_all_ntfs_filesystems(target):
    for fs, name, _ in iter_ntfs_filesystems(target):
        if name not in target.fs.mounts:
            target.fs.mount(name, fs)


def iter_esxi_filesystems(target):
    for mount, fs in target.fs.mounts.items():
        if not mount.startswith("/vmfs/volumes/"):
            continue

        uuid = mount[len("/vmfs/volumes/") :]  # strip /vmfs/volumes/
        name = None
        if fs.__fstype__ == "fat":
            name = fs.volume.name
        elif fs.__fstype__ == "vmfs":
            name = fs.vmfs.label

        yield uuid, name, fs


def register_module(*args, **kwargs):
    def wrapper(module_cls):
        name = module_cls.__name__

        if name in MODULES:
            raise ValueError(
                f"Module name is already registered: registration for {module_cls} conflicts with {MODULES[name]}"
            )

        desc = module_cls.DESC or name
        kwargs["help"] = f"acquire {desc}"
        kwargs["action"] = "store_true"
        kwargs["dest"] = name.lower()
        module_cls.__modname__ = name

        if not hasattr(module_cls, "__cli_args__"):
            module_cls.__cli_args__ = []
        module_cls.__cli_args__.append((args, kwargs))

        MODULES[name] = module_cls
        return module_cls

    return wrapper


def module_arg(*args, **kwargs):
    def wrapper(module_cls):
        if not hasattr(module_cls, "__cli_args__"):
            module_cls.__cli_args__ = []
        module_cls.__cli_args__.append((args, kwargs))
        return module_cls

    return wrapper


def local_module(cls):
    """A decorator that sets property `__local__` on a module class to mark it for local target only"""
    cls.__local__ = True
    return cls


class ExecutionOrder(enum.IntEnum):
    TOP = 0
    DEFAULT = 1
    BOTTOM = 2


class Module:
    DESC = None
    SPEC = []
    EXEC_ORDER = ExecutionOrder.DEFAULT

    @classmethod
    def run(cls, target, cli_args, collector):
        desc = cls.DESC or cls.__name__.lower()
        log.info("*** Acquiring %s", desc)

        collector.bind(cls)

        try:
            collector.collect(cls.SPEC)

            spec_ext = cls.get_spec_additions(target)
            if spec_ext:
                collector.collect(list(spec_ext))

            cls._run(target, collector)
        finally:
            collector.unbind()

    @classmethod
    def get_spec_additions(cls, target):
        pass

    @classmethod
    def _run(cls, target, collector):
        pass


@register_module("--sys")
@local_module
class Sys(Module):
    DESC = "Sysfs files (live systems only)"
    EXEC_ORDER = ExecutionOrder.BOTTOM

    @classmethod
    def _run(cls, target: Target, collector: Collector):
        if not Path("/sys").exists():
            log.error("/sys is unavailable! Skipping...")
            return

        spec = [("dir", "/sys")]

        sysfs = dir.DirectoryFilesystem(Path("/sys"))

        target.filesystems.add(sysfs)
        target.fs.mount("/sys", sysfs)

        collector.collect(spec, follow=False, volatile=True)


@register_module("--proc")
@local_module
class Proc(Module):
    DESC = "Procfs files (live systems only)"
    EXEC_ORDER = ExecutionOrder.BOTTOM

    @classmethod
    def _run(cls, target: Target, collector: Collector):
        if not Path("/proc").exists():
            log.error("/proc is unavailable! Skipping...")
            return

        spec = [("dir", "/proc")]
        procfs = dir.DirectoryFilesystem(Path("/proc"))

        target.filesystems.add(procfs)
        target.fs.mount("/proc", procfs)

        collector.collect(spec, follow=False, volatile=True)


@register_module("-n", "--ntfs")
class NTFS(Module):
    DESC = "NTFS filesystem metadata"

    @classmethod
    def _run(cls, target, collector):
        for fs, name, mountpoints in iter_ntfs_filesystems(target):
            log.info("Acquiring %s (%s)", fs, mountpoints)

            collector.collect_file(fs.path("$MFT"), outpath=name + "/$MFT")
            collector.collect_file(fs.path("$Boot"), outpath=name + "/$Boot")

            cls.collect_usnjrnl(collector, fs, name)
            cls.collect_ntfs_secure(collector, fs, name)

    @classmethod
    def collect_usnjrnl(cls, collector: Collector, fs, name: str) -> None:
        try:
            usnjrnl_path = fs.path("$Extend/$Usnjrnl:$J")
            entry = usnjrnl_path.get()
            journal = entry.open()

            i = 0
            while journal.runlist[i][0] is None:
                journal.seek(journal.runlist[i][1] * journal.block_size, io.SEEK_CUR)
                i += 1

            collector.output.write(
                f"{collector.base}/{name}/$Extend/$Usnjrnl:$J",
                journal,
                size=journal.size - journal.tell(),
                entry=entry,
            )
            collector.report.add_file_collected(cls.__name__, usnjrnl_path)
            result = "OK"
        except exceptions.FileNotFoundError:
            collector.report.add_file_missing(cls.__name__, usnjrnl_path)
            result = "File not found"
        except Exception as err:
            log.debug("Failed to acquire UsnJrnl", exc_info=True)
            collector.report.add_file_failed(cls.__name__, usnjrnl_path)
            result = repr(err)

        log.info("- Collecting file $Extend/$Usnjrnl:$J: %s", result)

    @classmethod
    def collect_ntfs_secure(cls, collector: Collector, fs, name: str) -> None:
        try:
            secure_path = fs.path("$Secure:$SDS")
            entry = secure_path.get()
            sds = entry.open()
            collector.output.write(
                f"{collector.base}/{name}/$Secure:$SDS",
                sds,
                size=sds.size,
                entry=entry,
            )
            collector.report.add_file_collected(cls.__name__, secure_path)
            result = "OK"
        except FileNotFoundError:
            collector.report.add_file_missing(cls.__name__, secure_path)
            result = "File not found"
        except Exception as err:
            log.debug("Failed to acquire SDS", exc_info=True)
            collector.report.add_file_failed(cls.__name__, secure_path)
            result = repr(err)

        log.info("- Collecting file $Secure:$SDS: %s", result)


@register_module("-r", "--registry")
class Registry(Module):
    DESC = "registry hives"
    HIVES = ["drivers", "sam", "security", "software", "system", "default"]
    SPEC = [
        ("dir", "sysvol/windows/system32/config/txr"),
        ("dir", "sysvol/windows/system32/config/regback"),
        ("glob", "sysvol/System Volume Information/_restore*/RP*/snapshot/_REGISTRY_*"),
        ("glob", "ntuser.dat*", from_user_home),
        ("glob", "AppData/Local/Microsoft/Windows/UsrClass.dat*", from_user_home),
        ("glob", "Local Settings/Application Data/Microsoft/Windows/UsrClass.dat*", from_user_home),
    ]

    @classmethod
    def get_spec_additions(cls, target):
        # Glob all hives to include e.g. .LOG files and .regtrans-ms files.
        files = []
        for hive in cls.HIVES:
            pattern = "sysvol/windows/system32/config/{}*".format(hive)
            for entry in target.fs.path().glob(pattern):
                if entry.is_file():
                    files.append(("file", entry))
        return files


@register_module("--netstat")
@local_module
class Netstat(Module):
    DESC = "netstat output"
    SPEC = [
        ("command", (["powershell.exe", "netstat", "-a", "-n", "-o"], "netstat")),
    ]
    EXEC_ORDER = ExecutionOrder.BOTTOM


@register_module("--win-processes")
@local_module
class WinProcesses(Module):
    DESC = "Windows process list"
    SPEC = [
        ("command", (["tasklist", "/V", "/fo", "csv"], "win-processes")),
    ]
    EXEC_ORDER = ExecutionOrder.BOTTOM


@register_module("--win-proc-env")
@local_module
class WinProcEnv(Module):
    DESC = "Process environment variables"
    SPEC = [
        (
            "command",
            (
                ["PowerShell", "-command", "Get-Process | ForEach-Object {$_.StartInfo.EnvironmentVariables}"],
                "win-process-env-vars",
            ),
        ),
    ]
    EXEC_ORDER = ExecutionOrder.BOTTOM


@register_module("--win-arp-cache")
@local_module
class WinArpCache(Module):
    DESC = "ARP Cache"
    EXEC_ORDER = ExecutionOrder.BOTTOM

    @classmethod
    def get_spec_additions(cls, target):
        if float(target.ntversion) < 6.2:
            commands = [
                # < Windows 10
                ("command", (["arp", "-av"], "win7-arp-cache")),
            ]
        else:
            commands = [
                # Windows 10+ (PowerShell)
                ("command", (["PowerShell", "Get-NetNeighbor"], "win10-arp-cache")),
            ]
        return commands


@register_module("--win-rdp-sessions")
@local_module
class WinRDPSessions(Module):
    DESC = "Windows Remote Desktop session information"
    EXEC_ORDER = ExecutionOrder.BOTTOM

    @classmethod
    def get_spec_additions(cls, target):
        # where.exe instead of where, just in case the client runs in PS instead of CMD
        # by default where hides qwinsta on 32-bit systems because qwinsta is only 64-bit, but with recursive /R search
        # we can still manage to find it and by passing the exact path Windows will launch a 64-bit process
        # on systems capable of doing that.
        qwinsta = subprocess.run(
            ["where.exe", "/R", os.environ["WINDIR"], "qwinsta.exe"], capture_output=True, text=True
        ).stdout.split("\n")[0]
        return [
            ("command", ([qwinsta, "/VM"], "win-rdp-sessions")),
        ]


@register_module("--winpmem")
@local_module
class WinMemDump(Module):
    DESC = "Windows full memory dump"
    EXEC_ORDER = ExecutionOrder.BOTTOM

    @classmethod
    def _run(cls, target, collector):
        winpmem_file_name = "winpmem.exe"
        winpmem_exec = shutil.which(winpmem_file_name)

        command_parts = [winpmem_exec, "-"]

        if winpmem_exec is None:
            command_parts.pop(0)
            command_parts.insert(0, winpmem_file_name)
            collector.report.add_command_failed(cls.__name__, command_parts)
            log.error(
                "- Failed to collect output from command `%s`, program %s not found",
                " ".join(command_parts),
                winpmem_file_name,
            )
            return

        else:
            log.info("- Collecting output from command `%s`", " ".join(command_parts))

            mem_dump_path = collector.output.path.with_name("winpmem")
            mem_dump_errors_path = mem_dump_path.with_name("winpmem.errors")

            output_base = collector.COMMAND_OUTPUT_BASE
            if collector.base:
                output_base = fsutil.join(collector.base, collector.COMMAND_OUTPUT_BASE)

            mem_dump_output_path = fsutil.join(output_base, mem_dump_path.name)
            mem_dump_errors_output_path = fsutil.join(output_base, mem_dump_errors_path.name)

            with mem_dump_path.open(mode="wb") as mem_dump_fh:
                with mem_dump_errors_path.open(mode="wb") as mem_dump_errors_fh:
                    try:
                        # The shell parameter must be set to False, as otherwise the
                        # output from stdout is not piped into the filehandle.
                        # The check parameter must be set to False, as winpmem.exe
                        # always seems to exit with an error code, even on success.
                        subprocess.run(
                            bufsize=0,
                            args=command_parts,
                            stdout=mem_dump_fh,
                            stderr=mem_dump_errors_fh,
                            shell=False,
                            check=False,
                        )

                    except Exception:
                        collector.report.add_command_failed(cls.__name__, command_parts)
                        log.error(
                            "- Failed to collect output from command `%s`",
                            " ".join(command_parts),
                            exc_info=True,
                        )
                        return

            collector.output.write_entry(mem_dump_output_path, entry=mem_dump_path)
            collector.output.write_entry(mem_dump_errors_output_path, entry=mem_dump_errors_path)
            collector.report.add_command_collected(cls.__name__, command_parts)
            mem_dump_path.unlink()
            mem_dump_errors_path.unlink()


@register_module("-e", "--eventlogs")
class EventLogs(Module):
    DESC = "event logs"

    @classmethod
    def get_spec_additions(cls, target):
        spec = set()
        evt_log_paths = evt.EvtPlugin(target).get_logs(filename_glob="*.evt")
        for path in evt_log_paths:
            spec.add(("file", path))
        evtx_log_paths = evtx.EvtxPlugin(target).get_logs(filename_glob="*.evtx")
        for path in evtx_log_paths:
            spec.add(("file", path))
        return spec


@register_module("-t", "--tasks")
class Tasks(Module):
    SPEC = [
        ("dir", "sysvol/windows/tasks"),
        ("dir", "sysvol/windows/system32/tasks"),
        ("dir", "sysvol/windows/syswow64/tasks"),
    ]


@register_module("-nt", "--ntds")
class NTDS(Module):
    SPEC = [
        ("dir", "sysvol/windows/NTDS"),
    ]

    @classmethod
    def get_spec_additions(cls, target):
        spec = set()
        key = "HKLM\\SYSTEM\\CurrentControlSet\\services\\NTDS\\Parameters"
        values = [
            ("dir", "DSA Working Directory"),
            ("file", "DSA Database File"),
            ("file", "Database backup path"),
            ("dir", "Database log files path"),
        ]
        for reg_key in target.registry.iterkeys(key):
            for collect_type, value in values:
                path = reg_key.value(value).value
                spec.add((collect_type, path))
        return spec


@register_module("--etl")
class ETL(Module):
    DESC = "interesting ETL files"
    SPEC = [
        ("glob", "sysvol/Windows/System32/WDI/LogFiles/*.etl"),
    ]


@register_module("--recents")
class Recents(Module):
    DESC = "Windows recently used files artifacts"
    SPEC = [
        ("dir", "AppData/Roaming/Microsoft/Windows/Recent", from_user_home),
        ("dir", "AppData/Roaming/Microsoft/Office/Recent", from_user_home),
        ("glob", "Recent/*.lnk", from_user_home),
    ]


@register_module("--recyclebin")
class RecycleBin(Module):
    DESC = "recycle bin metadata"

    @classmethod
    def _run(cls, target, collector):
        for fs, name, mountpoints in iter_ntfs_filesystems(target):
            log.info("Acquiring recycle bin metadata from %s (%s)", fs, mountpoints)

            patterns = ["$Recycle.bin/**/$I*", "Recycler/*/INFO2", "Recycled/INFO2"]
            for pattern in patterns:
                for entry in fs.path().glob(pattern):
                    collector.collect_file(entry, outpath=fsutil.join(name, str(entry)))


@register_module("--drivers")
class Drivers(Module):
    DESC = "installed drivers"
    SPEC = [
        ("glob", "sysvol/windows/system32/drivers/*.sys"),
    ]


@register_module("--exchange")
class Exchange(Module):
    DESC = "interesting Exchange configuration files"

    @classmethod
    def get_spec_additions(cls, target):
        spec = set()

        key = "HKLM\\SOFTWARE\\Microsoft\\ExchangeServer"
        for reg_key in target.registry.iterkeys(key):
            for subkey in reg_key.subkeys():
                try:
                    setup_key = subkey.subkey("Setup")
                    install_path = setup_key.value("MsiInstallPath").value
                    spec.update(
                        [
                            (
                                "file",
                                f"{install_path}\\TransportRoles\\Agents\\agents.config",
                            ),
                            (
                                "dir",
                                f"{install_path}\\Logging\\Ews",
                            ),
                            (
                                "dir",
                                f"{install_path}\\Logging\\CmdletInfra\\Powershell-Proxy\\Cmdlet",
                            ),
                            (
                                "dir",
                                f"{install_path}\\TransportRoles\\Logs",
                            ),
                        ]
                    )
                except Exception:
                    pass
        return spec


@register_module("--iis")
class IIS(Module):
    DESC = "IIS logs"

    @classmethod
    def get_spec_additions(cls, target):
        spec = set(
            [
                ("glob", "sysvol\\Windows\\System32\\LogFiles\\W3SVC*\\*.log"),
                ("glob", "sysvol\\Windows.old\\Windows\\System32\\LogFiles\\W3SVC*\\*.log"),
                ("glob", "sysvol\\inetpub\\logs\\LogFiles\\*.log"),
                ("glob", "sysvol\\inetpub\\logs\\LogFiles\\W3SVC*\\*.log"),
                ("glob", "sysvol\\Resources\\Directory\\*\\LogFiles\\Web\\W3SVC*\\*.log"),
            ]
        )
        iis_plugin = iis.IISLogsPlugin(target)
        spec.update([("file", log_path) for _, log_path in iis_plugin.iter_log_format_path_pairs()])
        return spec


@register_module("--prefetch")
class Prefetch(Module):
    DESC = "Windows Prefetch files"
    SPEC = [
        ("dir", "sysvol/windows/prefetch"),
    ]


@register_module("--appcompat")
class Appcompat(Module):
    DESC = "Windows Amcache and RecentFileCache"
    SPEC = [
        ("dir", "sysvol/windows/appcompat"),
    ]


@register_module("--pca")
class PCA(Module):
    DESC = "Windows Program Compatibility Assistant"
    SPEC = [
        ("dir", "sysvol/windows/pca"),
    ]


@register_module("--syscache")
class Syscache(Module):
    DESC = "Windows Syscache hive and log files"
    SPEC = [
        ("file", "sysvol/System Volume Information/Syscache.hve"),
        ("glob", "sysvol/System Volume Information/Syscache.hve.LOG*"),
    ]


@register_module("--win-notifications")
class WindowsNotifications(Module):
    DESC = "Windows Push Notifications Database files."
    SPEC = [
        # Old Win7/Win10 version of the file
        ("file", "AppData/Local/Microsoft/Windows/Notifications/appdb.dat", from_user_home),
        # New version of the file
        ("file", "AppData/Local/Microsoft/Windows/Notifications/wpndatabase.db", from_user_home),
    ]


@register_module("--bits")
class BITS(Module):
    DESC = "Background Intelligent Transfer Service (BITS) queue/log DB"
    SPEC = [
        # Pre-Win10 the BITS DB files are called qmgr[01].dat, in Win10 it is
        # called qmgr.db and its transaction logs edb.log and edb.log[0-2]
        # Win 2000/XP/2003 path
        # (basically: \%ALLUSERSPROFILE%\Application Data\Microsoft\...)
        ("glob", "sysvol/Documents and Settings/All Users/Application Data/Microsoft/Network/Downloader/qmgr*.dat"),
        # Win Vista and higher path
        # (basically: \%ALLUSERSPROFILE%\Microsoft\...; %ALLUSERSPROFILE% == %PROGRAMDATA%)
        ("glob", "sysvol/ProgramData/Microsoft/Network/Downloader/qmgr*.dat"),
        # Win 10 files
        ("file", "sysvol/ProgramData/Microsoft/Network/Downloader/qmgr.db"),
        ("glob", "sysvol/ProgramData/Microsoft/Network/Downloader/edb.log*"),
    ]


@register_module("--wbem")
class WBEM(Module):
    DESC = "Windows WBEM (WMI) database files"
    SPEC = [
        ("dir", "sysvol/windows/system32/wbem/Repository"),
    ]


@register_module("--dhcp")
class DHCP(Module):
    DESC = "Windows Server DHCP files"

    @classmethod
    def get_spec_additions(cls, target):
        spec = set()
        key = "HKLM\\SYSTEM\\CurrentControlSet\\Services\\DhcpServer\\Parameters"
        for reg_key in target.registry.iterkeys(key):
            spec.add(("dir", reg_key.value("DatabasePath").value))
        return spec


@register_module("--dns")
class DNS(Module):
    DESC = "Windows Server DNS files"
    SPEC = [
        ("glob", "sysvol/windows/system32/config/netlogon.*"),
        ("dir", "sysvol/windows/system32/dns"),
    ]


@register_module("--win-dns-cache")
@local_module
class WinDnsClientCache(Module):
    DESC = "The contents of Windows DNS client cache"
    SPEC = [
        (
            "command",
            # Powershell.exe understands a subcommand passed as single string parameter,
            # no need to split the subcommand in parts.
            (
                ["powershell.exe", "-Command", "Get-DnsClientCache | ConvertTo-Csv -NoTypeInformation"],
                "get-dnsclientcache",
            ),
        ),
    ]
    EXEC_ORDER = ExecutionOrder.BOTTOM


@register_module("--powershell")
class PowerShell(Module):
    DESC = "Windows PowerShell Artefacts"
    SPEC = [
        ("dir", "AppData/Roaming/Microsoft/Windows/PowerShell", from_user_home),
    ]


@register_module("--thumbnail-cache")
class ThumbnailCache(Module):
    DESC = "Windows thumbnail db artifacts"
    SPEC = [
        ("glob", "AppData/Local/Microsoft/Windows/Explorer/thumbcache_*", from_user_home),
    ]


@register_module("--misc")
class Misc(Module):
    DESC = "miscellaneous Windows artefacts"
    SPEC = [
        ("file", "sysvol/windows/PFRO.log"),
        ("file", "sysvol/windows/setupapi.log"),
        ("file", "sysvol/windows/setupapidev.log"),
        ("glob", "sysvol/windows/inf/setupapi*.log"),
        ("glob", "sysvol/system32/logfiles/*/*.txt"),
        ("dir", "sysvol/windows/system32/sru"),
        ("dir", "sysvol/windows/system32/drivers/etc"),
        ("dir", "sysvol/Windows/System32/WDI/LogFiles/StartupInfo"),
        ("dir", "sysvol/windows/sysvol/domain/policies/"),
        ("dir", "sysvol/windows/system32/GroupPolicy/DataStore/"),
        ("dir", "sysvol/ProgramData/Microsoft/Group Policy/History/"),
        ("glob", "sysvol/Windows/System32/LogFiles/SUM/*.mdb"),
    ]


@register_module("--av")
class AV(Module):
    DESC = "various antivirus logs"
    SPEC = [
        # AVG
        ("dir", "sysvol/Documents and Settings/All Users/Application Data/AVG/Antivirus/log"),
        ("dir", "sysvol/Documents and Settings/All Users/Application Data/AVG/Antivirus/report"),
        ("dir", "sysvol/ProgramData/AVG/Antivirus/log"),
        ("dir", "sysvol/ProgramData/AVG/Antivirus/report"),
        # Avast
        ("dir", "sysvol/Documents And Settings/All Users/Application Data/Avast Software/Avast/Log"),
        ("dir", "sysvol/ProgramData/Avast Software/Avast/Log"),
        ("dir", "Avast Software/Avast/Log", from_user_home),
        ("file", "sysvol/ProgramData/Avast Software/Avast/Chest/index.xml"),
        # Avira
        ("dir", "sysvol/ProgramData/Avira/Antivirus/LOGFILES"),
        # Bitdefender
        ("dir", "sysvol/ProgramData/Bitdefender/Endpoint Security/Logs"),
        ("dir", "sysvol/ProgramData/Bitdefender/Desktop/Profiles/Logs"),
        ("glob", "sysvol/Program Files*/Bitdefender*/*"),
        # ComboFix
        ("file", "sysvol/ComboFix.txt"),
        # Cybereason
        ("dir", "sysvol/ProgramData/crs1/Logs"),
        ("dir", "sysvol/ProgramData/apv2/Logs"),
        ("dir", "sysvol/ProgramData/crb1/Logs"),
        # ESET
        ("dir", "sysvol/Documents and Settings/All Users/Application Data/ESET/ESET NOD32 Antivirus/Logs"),
        ("dir", "sysvol/ProgramData/ESET/ESET NOD32 Antivirus/Logs"),
        # Emsisoft
        ("glob", "sysvol/ProgramData/Emsisoft/Reports/scan*.txt"),
        # F-Secure
        ("dir", "sysvol/ProgramData/F-Secure/Log"),
        ("dir", "sysvol/Users*/AppData/Local/F-Secure/Log"),
        ("dir", "sysvol/ProgramData/F-Secure/Antivirus/ScheduledScanReports"),
        # HitmanPro
        ("dir", "sysvol/ProgramData/HitmanPro/Logs"),
        ("dir", "sysvol/ProgramData/HitmanPro.Alert/Logs"),
        ("file", "sysvol/ProgramData/HitmanPro.Alert/excalibur.db"),
        ("glob", "sysvol/ProgramData/HitmanPro/Quarantine"),
        # Malwarebytes
        ("glob", "sysvol/ProgramData/Malwarebytes/Malwarebytes Anti-Malware/Logs/mbam-log-*.xml"),
        ("glob", "sysvol/ProgramData/Malwarebytes/MBAMService/logs/mbamservice.log*"),
        ("dir", "sysvol/Users*/AppData/Roaming/Malwarebytes/Malwarebytes Anti-Malware/Logs"),
        ("dir", "sysvol/ProgramData/Malwarebytes/MBAMService/ScanResults"),
        # McAfee
        ("dir", "sysvol/Users/All Users/Application Data/McAfee/DesktopProtection"),
        ("dir", "sysvol/ProgramData/McAfee/DesktopProtection"),
        ("dir", "sysvol/ProgramData/McAfee/Endpoint Security/Logs"),
        ("dir", "sysvol/ProgramData/McAfee/Endpoint Security/Logs_Old"),
        ("dir", "sysvol/ProgramData/Mcafee/VirusScan"),
        ("dir", "sysvol/ProgramData/McAfee/Endpoint Security/Logs"),
        ("dir", "sysvol/ProgramData/McAfee/MSC/Logs"),
        # RogueKiller
        ("glob", "sysvol/ProgramData/RogueKiller/logs/AdliceReport_*.json"),
        # SUPERAntiSpyware
        ("dir", "sysvol/Users*/AppData/Roaming/SUPERAntiSpyware/Logs"),
        # SecureAge
        ("dir", "sysvol/ProgramData/SecureAge Technology/SecureAge/log"),
        # SentinelOne
        ("dir", "sysvol/programdata/sentinel/logs"),
        # Sophos
        ("glob", "sysvol/Documents and Settings/All Users/Application Data/Sophos/Sophos */Logs"),
        ("glob", "sysvol/ProgramData/Sophos/Sophos */Logs"),
        # Symantec
        (
            "dir",
            "sysvol/Documents and Settings/All Users/Application Data/Symantec/Symantec Endpoint Protection/Logs/AV",
        ),
        ("glob", "sysvol/ProgramData/Symantec/Symantec Endpoint Protection/*/Data/Logs"),
        ("dir", "AppData/Local/Symantec/Symantec Endpoint Protection/Logs", from_user_home),
        ("dir", "sysvol/Windows/System32/winevt/logs/Symantec Endpoint Protection Client.evtx"),
        ("glob", "sysvol/ProgramData/Symantec/Symantec Endpoint Protection/*/Data/CmnClnt/ccSubSDK"),
        ("glob", "sysvol/ProgramData/Symantec/Symantec Endpoint Protection/*/Data/registrationInfo.xml"),
        # TotalAV
        ("glob", "sysvol/Program Files*/TotalAV/logs"),
        ("dir", "sysvol/ProgramData/TotalAV/logs"),
        # VIPRE
        ("dir", "sysvol/ProgramData/VIPRE Business Agent/Logs"),
        ("dir", "AppData/Roaming/VIPRE Business", from_user_home),
        ("dir", "AppData/Roaming/GFI Software/AntiMalware/Logs", from_user_home),
        ("dir", "AppData/Roaming/Sunbelt Software/AntiMalware/Logs", from_user_home),
        # Webroot
        ("file", "sysvol/ProgramData/WRData/WRLog.log"),
        # Microsoft Windows Defender
        ("dir", "sysvol/ProgramData/Microsoft/Microsoft AntiMalware/Support"),
        ("glob", "sysvol/Windows/System32/winevt/Logs/Microsoft-Windows-Windows Defender*.evtx"),
        ("dir", "sysvol/ProgramData/Microsoft/Windows Defender/Support"),
        ("file", "sysvol/Windows/Temp/MpCmdRun.log"),
        ("file", "sysvol/Windows.old/Windows/Temp/MpCmdRun.log"),
    ]


@register_module("--quarantined")
class QuarantinedFiles(Module):
    DESC = "files quarantined by various antivirus products"
    SPEC = [
        # Microsoft Defender
        # https://knez.github.io/posts/how-to-extract-quarantine-files-from-windows-defender/
        ("dir", "sysvol/ProgramData/Microsoft/Windows Defender/Quarantine"),
        # Symantec Endpoint Protection
        (
            "dir",
            "sysvol/Documents and Settings/All Users/Application Data/Symantec/Symantec Endpoint Protection/Quarantine",
        ),
        ("glob", "sysvol/ProgramData/Symantec/Symantec Endpoint Protection/*/Data/Quarantine"),
        # Trend Micro
        # https://secret.inf.ufpr.br/papers/marcus_av_handson.pdf
        ("dir", "sysvol/ProgramData/Trend Micro/AMSP/quarantine"),
        # McAfee
        ("dir", "sysvol/Quarantine"),
        ("dir", "sysvol/ProgramData/McAfee/VirusScan/Quarantine"),
        # Sophos
        ("glob", "sysvol/ProgramData/Sophos/Sophos/*/Quarantine"),
        ("glob", "sysvol/ProgramData/Sophos/Sophos */INFECTED"),
        ("dir", "sysvol/ProgramData/Sophos/Safestore"),
        # HitmanPRO
        ("dir", "sysvol/ProgramData/HitmanPro/Quarantine"),
    ]


@register_module("--history")
class History(Module):
    DESC = "browser history from IE, Edge, Firefox, and Chrome"

    SPEC = [
        # IE
        ("file", "Cookies/index.dat", from_user_home),
        ("file", "Local Settings/History/History.IE5/index.dat", from_user_home),
        ("glob", "Local Settings/History/History.IE5/MSHist*/index.dat", from_user_home),
        ("file", "Local Settings/Temporary Internet Files/Content.IE5/index.dat", from_user_home),
        ("file", "Local Settings/Application Data/Microsoft/Feeds Cache/index.dat", from_user_home),
        ("dir", "AppData/Local/Microsoft/Internet Explorer/Recovery", from_user_home),
        ("file", "AppData/Local/Microsoft/Windows/History/History.IE5/index.dat", from_user_home),
        (
            "file",
            "AppData/Local/Microsoft/Windows/History/History.IE5/MSHist*/index.dat",
            from_user_home,
        ),
        (
            "file",
            "AppData/Local/Microsoft/Windows/History/Low/History.IE5/index.dat",
            from_user_home,
        ),
        (
            "file",
            "AppData/Local/Microsoft/Windows/History/Low/History.IE5/MSHist*/index.dat",
            from_user_home,
        ),
        ("dir", "AppData/Local/Microsoft/Windows/INetCookies", from_user_home),
        (
            "file",
            "AppData/Local/Microsoft/Windows/Temporary Internet Files/Content.IE5/index.dat",
            from_user_home,
        ),
        (
            "file",
            "AppData/Local/Microsoft/Windows/Temporary Internet Files/Low/Content.IE5/index.dat",
            from_user_home,
        ),
        ("glob", "AppData/Local/Microsoft/Windows/WebCache/*.dat", from_user_home),
        ("file", "AppData/Roaming/Microsoft/Windows/Cookies/index.dat", from_user_home),
        ("file", "AppData/Roaming/Microsoft/Windows/Cookies/Low/index.dat", from_user_home),
        ("file", "AppData/Roaming/Microsoft/Windows/IEDownloadHistory/index.dat", from_user_home),
        # Chrome
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Bookmarks", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Favicons", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/History", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Login Data", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Login Data For Account", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Shortcuts", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Top Sites", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Web Data", from_user_home),
        # Chrome - Legacy
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Current Session", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Current Tabs", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Archived History", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Last Session", from_user_home),
        ("glob", "AppData/Local/Google/Chrom*/User Data/*/Last Tabs", from_user_home),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Bookmarks",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Favicons",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/History",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Login Data",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Login Data For Account",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Shortcuts",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Top Sites",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Web Data",
            from_user_home,
        ),
        # Chrome - Legacy
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Current Session",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Current Tabs",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Archived History",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Last Session",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Google/Chrom*/User Data/*/Last Tabs",
            from_user_home,
        ),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Bookmarks"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Favicons"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/History"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Login Data"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Login Data For Account"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Shortcuts"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Top Sites"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Web Data"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Bookmarks"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Favicons"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/History"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Login Data"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Login Data For Account"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Shortcuts"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Top Sites"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Web Data"),
        # Chrome - Legacy
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Current Session"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Current Tabs"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Archived History"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Last Session"),
        ("glob", "/Users/*/Library/Application Support/Google/Chrome/*/Last Tabs"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Current Session"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Current Tabs"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Archived History"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Last Session"),
        ("glob", "/Users/*/Library/Application Support/Chromium/*/Last Tabs"),
        # Chrome - RHEL/Ubuntu - DNF
        ("glob", ".config/google-chrome/*/Bookmarks", from_user_home),
        ("glob", ".config/google-chrome/*/Favicons", from_user_home),
        ("glob", ".config/google-chrome/*/History", from_user_home),
        ("glob", ".config/google-chrome/*/Login Data", from_user_home),
        ("glob", ".config/google-chrome/*/Login Data For Account", from_user_home),
        ("glob", ".config/google-chrome/*/Shortcuts", from_user_home),
        ("glob", ".config/google-chrome/*/Top Sites", from_user_home),
        ("glob", ".config/google-chrome/*/Web Data", from_user_home),
        # Chrome - RHEL/Ubuntu - Flatpak
        ("glob", ".var/app/com.google.Chrome/config/google-chrome/*/Bookmarks", from_user_home),
        ("glob", ".var/app/com.google.Chrome/config/google-chrome/*/Favicons", from_user_home),
        ("glob", ".var/app/com.google.Chrome/config/google-chrome/*/History", from_user_home),
        ("glob", ".var/app/com.google.Chrome/config/google-chrome/*/Login Data", from_user_home),
        ("glob", ".var/app/com.google.Chrome/config/google-chrome/*/Login Data For Account", from_user_home),
        ("glob", ".var/app/com.google.Chrome/config/google-chrome/*/Shortcuts", from_user_home),
        ("glob", ".var/app/com.google.Chrome/config/google-chrome/*/Top Sites", from_user_home),
        ("glob", ".var/app/com.google.Chrome/config/google-chrome/*/Web Data", from_user_home),
        # Chromium - RHEL/Ubuntu - DNF/apt
        ("glob", ".config/chromium/*/Bookmarks", from_user_home),
        ("glob", ".config/chromium/*/Favicons", from_user_home),
        ("glob", ".config/chromium/*/History", from_user_home),
        ("glob", ".config/chromium/*/Login Data", from_user_home),
        ("glob", ".config/chromium/*/Login Data For Account", from_user_home),
        ("glob", ".config/chromium/*/Shortcuts", from_user_home),
        ("glob", ".config/chromium/*/Top Sites", from_user_home),
        ("glob", ".config/chromium/*/Web Data", from_user_home),
        # Chromium - RHEL/Ubuntu - Flatpak
        ("glob", ".var/app/org.chromium.Chromium/config/chromium/*/Bookmarks", from_user_home),
        ("glob", ".var/app/org.chromium.Chromium/config/chromium/*/Favicons", from_user_home),
        ("glob", ".var/app/org.chromium.Chromium/config/chromium/*/History", from_user_home),
        ("glob", ".var/app/org.chromium.Chromium/config/chromium/*/Login Data", from_user_home),
        ("glob", ".var/app/org.chromium.Chromium/config/chromium/*/Login Data For Account", from_user_home),
        ("glob", ".var/app/org.chromium.Chromium/config/chromium/*/Shortcuts", from_user_home),
        ("glob", ".var/app/org.chromium.Chromium/config/chromium/*/Top Sites", from_user_home),
        ("glob", ".var/app/org.chromium.Chromium/config/chromium/*/Web Data", from_user_home),
        # Chromium - RHEL/Ubuntu - snap
        ("glob", "snap/chromium/common/chromium/*/Bookmarks", from_user_home),
        ("glob", "snap/chromium/common/chromium/*/Favicons", from_user_home),
        ("glob", "snap/chromium/common/chromium/*/History", from_user_home),
        ("glob", "snap/chromium/common/chromium/*/Login Data", from_user_home),
        ("glob", "snap/chromium/common/chromium/*/Login Data For Account", from_user_home),
        ("glob", "snap/chromium/common/chromium/*/Shortcuts", from_user_home),
        ("glob", "snap/chromium/common/chromium/*/Top Sites", from_user_home),
        ("glob", "snap/chromium/common/chromium/*/Web Data", from_user_home),
        # Edge
        ("glob", "AppData/Local/Microsoft/Edge/User Data/*/Bookmarks", from_user_home),
        ("glob", "AppData/Local/Microsoft/Edge/User Data/*/Extension Cookies", from_user_home),
        ("glob", "AppData/Local/Microsoft/Edge/User Data/*/Favicons", from_user_home),
        ("glob", "AppData/Local/Microsoft/Edge/User Data/*/History", from_user_home),
        ("glob", "AppData/Local/Microsoft/Edge/User Data/*/Login Data", from_user_home),
        ("glob", "AppData/Local/Microsoft/Edge/User Data/*/Media History", from_user_home),
        ("glob", "AppData/Local/Microsoft/Edge/User Data/*/Shortcuts", from_user_home),
        ("glob", "AppData/Local/Microsoft/Edge/User Data/*/Top Sites", from_user_home),
        ("glob", "AppData/Local/Microsoft/Edge/User Data/*/Web Data", from_user_home),
        (
            "glob",
            "Local Settings/Application Data/Microsoft/Edge/User Data/*/Bookmarks",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Microsoft/Edge/User Data/*/Extension Cookies",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Microsoft/Edge/User Data/*/Favicons",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Microsoft/Edge/User Data/*/History",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Microsoft/Edge/User Data/*/Login Data",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Microsoft/Edge/User Data/*/Media History",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Microsoft/Edge/User Data/*/Shortcuts",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Microsoft/Edge/User Data/*/Top Sites",
            from_user_home,
        ),
        (
            "glob",
            "Local Settings/Application Data/Microsoft/Edge/User Data/*/Web Data",
            from_user_home,
        ),
        ("glob", "/Users/*/Library/Application Support/Microsoft Edge/*/Bookmarks"),
        ("glob", "/Users/*/Library/Application Support/Microsoft Edge/*/Extension Cookies"),
        ("glob", "/Users/*/Library/Application Support/Microsoft Edge/*/Favicons"),
        ("glob", "/Users/*/Library/Application Support/Microsoft Edge/*/History"),
        ("glob", "/Users/*/Library/Application Support/Microsoft Edge/*/Login Data"),
        ("glob", "/Users/*/Library/Application Support/Microsoft Edge/*/Media History"),
        ("glob", "/Users/*/Library/Application Support/Microsoft Edge/*/Shortcuts"),
        ("glob", "/Users/*/Library/Application Support/Microsoft Edge/*/Top Sites"),
        ("glob", "/Users/*/Library/Application Support/Microsoft Edge/*/Web Data"),
        # Edge - RHEL/Ubuntu - DNF/apt
        ("glob", ".config/microsoft-edge/*/Bookmarks", from_user_home),
        ("glob", ".config/microsoft-edge/*/Favicons", from_user_home),
        ("glob", ".config/microsoft-edge/*/History", from_user_home),
        ("glob", ".config/microsoft-edge/*/Login Data", from_user_home),
        ("glob", ".config/microsoft-edge/*/Login Data For Account", from_user_home),
        ("glob", ".config/microsoft-edge/*/Shortcuts", from_user_home),
        ("glob", ".config/microsoft-edge/*/Top Sites", from_user_home),
        ("glob", ".config/microsoft-edge/*/Web Data", from_user_home),
        # Edge - RHEL/Ubuntu - Flatpak
        ("glob", ".var/app/com.microsoft.Edge/config/microsoft-edge/*/Bookmarks", from_user_home),
        ("glob", ".var/app/com.microsoft.Edge/config/microsoft-edge/*/Favicons", from_user_home),
        ("glob", ".var/app/com.microsoft.Edge/config/microsoft-edge/*/History", from_user_home),
        ("glob", ".var/app/com.microsoft.Edge/config/microsoft-edge/*/Login Data", from_user_home),
        ("glob", ".var/app/com.microsoft.Edge/config/microsoft-edge/*/Login Data For Account", from_user_home),
        ("glob", ".var/app/com.microsoft.Edge/config/microsoft-edge/*/Shortcuts", from_user_home),
        ("glob", ".var/app/com.microsoft.Edge/config/microsoft-edge/*/Top Sites", from_user_home),
        ("glob", ".var/app/com.microsoft.Edge/config/microsoft-edge/*/Web Data", from_user_home),
        # Firefox - Windows
        ("glob", "AppData/Local/Mozilla/Firefox/Profiles/*/*.sqlite*", from_user_home),
        ("glob", "AppData/Roaming/Mozilla/Firefox/Profiles/*/*.sqlite*", from_user_home),
        ("glob", "Application Data/Mozilla/Firefox/Profiles/*/*.sqlite*", from_user_home),
        # Firefox - macOS
        ("glob", "/Users/*/Library/Application Support/Firefox/Profiles/*/*.sqlite*"),
        # Firefox - RHEL/Ubuntu - Flatpak
        ("glob", ".var/app/org.mozilla.firefox/.mozilla/firefox/*/*.sqlite", from_user_home),
        # Firefox - RHEL/Ubuntu - DNF/apt
        ("glob", ".mozilla/firefox/*/*.sqlite", from_user_home),
        # Firefox - RHEL/Ubuntu - snap
        ("glob", "snap/firefox/common/.mozilla/firefox/*/*.sqlite", from_user_home),
        # Safari - macOS
        ("glob", "/Users/*/Library/Safari/Bookmarks.plist"),
        ("glob", "/Users/*/Library/Safari/Downloads.plist"),
        ("glob", "/Users/*/Library/Safari/Extensions/Extensions.plist"),
        ("glob", "/Users/*/Library/Safari/History.*"),
        ("glob", "/Users/*/Library/Safari/LastSession.plist"),
        ("glob", "/Users/*/Library/Caches/com.apple.Safari/Cache.db"),
    ]


@register_module("--remoteaccess")
class RemoteAccess(Module):
    DESC = "common remote access tools' log files"
    SPEC = [
        # teamviewer
        ("glob", "sysvol/Program Files/TeamViewer/*.log"),
        ("glob", "sysvol/Program Files (x86)/TeamViewer/*.log"),
        ("glob", "AppData/Roaming/TeamViewer/*.log", from_user_home),
        # anydesk
        ("dir", "sysvol/ProgramData/AnyDesk"),
        ("glob", "AppData/Roaming/AnyDesk/*.trace", from_user_home),
        # zoho
        ("dir", "sysvol/ProgramData/ZohoMeeting/log"),
        ("dir", "AppData/Local/ZohoMeeting/log", from_user_home),
        # realvnc
        ("file", "sysvol/ProgramData/RealVNC-Service/vncserver.log"),
        ("file", "AppData/Local/RealVNC/vncserver.log", from_user_home),
        # tightvnc
        ("dir", "sysvol/ProgramData/TightVNC/Server/Logs"),
        # Remote desktop cache files
        ("dir", "AppData/Local/Microsoft/Terminal Server Client/Cache", from_user_home),
    ]


@register_module("--wer")
class WER(Module):
    DESC = "WER (Windows Error Reporting) related files"

    @classmethod
    def get_spec_additions(cls, target):
        spec = set()

        for wer_dir in itertools.chain(
            ["sysvol/ProgramData/Microsoft/Windows/WER"],
            from_user_home(target, "AppData/Local/Microsoft/Windows/WER"),
        ):
            for path in target.fs.path(wer_dir).rglob("*"):
                if not path.is_file():
                    continue

                if path.stat().st_size >= (1024 * 1024 * 1024):  # 1GB
                    log.debug("Skipping WER file because it exceeds 1GB: %s", path)
                    continue

                spec.add(("file", path))

        return spec


@register_module("--etc")
class Etc(Module):
    SPEC = [
        # In OS-X /etc is a symlink to /private/etc. To prevent collecting
        # duplicates, we only use the /etc directory here.
        ("dir", "/etc"),
        ("dir", "/usr/local/etc"),
    ]


@register_module("--boot")
class Boot(Module):
    SPEC = [
        ("glob", "/boot/config*"),
        ("glob", "/boot/efi*"),
        ("glob", "/boot/grub*"),
        ("glob", "/boot/init*"),
    ]


@register_module("--home")
class Home(Module):
    SPEC = [
        ("glob", "/root/.*[akz]sh*"),
        ("dir", "/root/.config"),
        ("glob", "/home/*/.*[akz]sh*"),
        ("glob", "/home/*/.config"),
        ("glob", "/home/*/*/.*[akz]sh*"),
        ("glob", "/home/*/*/.config"),
        # OS-X home (aka /Users)
        ("glob", "/Users/*/.*[akz]sh*"),
        ("glob", "/Users/*/.bash_sessions/*"),
        ("glob", "/Users/*/Library/LaunchAgents/*"),
        ("glob", "/Users/*/Library/Logs/*"),
        ("glob", "/Users/*/Preferences/*"),
        ("glob", "/Users/*/Library/Preferences/*"),
    ]


@register_module("--var")
class Var(Module):
    SPEC = [
        # In OS-X /var is a symlink to /private/var. To prevent collecting
        # duplicates, we only use the /var directory here.
        ("dir", "/var/log"),
        ("dir", "/var/spool/at"),
        ("dir", "/var/spool/cron"),
        ("dir", "/var/spool/anacron"),
        ("dir", "/var/lib/dpkg/status"),
        ("dir", "/var/lib/rpm"),
        ("dir", "/var/db"),
        ("dir", "/var/audit"),
        ("dir", "/var/cron"),
        ("dir", "/var/run"),
        # some OS-X specific files
        ("dir", "/private/var/at"),
        ("dir", "/private/var/db/diagnostics"),
        ("dir", "/private/var/db/uuidtext"),
        ("file", "/private/var/vm/sleepimage"),
        ("glob", "/private/var/vm/swapfile*"),
        ("glob", "/private/var/folders/*/*/0/com.apple.notificationcenter/*/*"),
        # user specific cron on OS-X
        ("dir", "/usr/lib/cron"),
    ]


@register_module("--bsd")
class BSD(Module):
    SPEC = [
        ("file", "/bin/freebsd-version"),
        ("dir", "/usr/ports"),
    ]


@register_module("--osx")
class OSX(Module):
    DESC = "OS-X specific files and directories"
    SPEC = [
        # filesystem events
        ("dir", "/.fseventsd"),
        # kernel extensions
        ("dir", "/Library/Extensions"),
        ("dir", "/System/Library/Extensions"),
        # logs
        ("dir", "/Library/Logs"),
        # autorun locations
        ("dir", "/Library/LaunchAgents"),
        ("dir", "/Library/LaunchDaemons"),
        ("dir", "/Library/StartupItems"),
        ("dir", "/System/Library/LaunchAgents"),
        ("dir", "/System/Library/LaunchDaemons"),
        ("dir", "/System/Library/StartupItems"),
        # installed software
        ("dir", "/Library/Receipts/InstallHistory.plist"),
        ("file", "/System/Library/CoreServices/SystemVersion.plist"),
        # system preferences
        ("dir", "/Library/Preferences"),
    ]


@register_module("--bootbanks")
class Bootbanks(Module):
    DESC = "ESXi bootbanks"

    @classmethod
    def _run(cls, target, collector):
        # Both ESXi 6 and 7 compatible
        boot_dirs = {
            "boot": "BOOT",
            "bootbank": "BOOTBANK1",
            "altbootbank": "BOOTBANK2",
        }
        boot_fs = []

        for boot_dir, boot_vol in boot_dirs.items():
            dir_path = target.fs.path(boot_dir)
            if dir_path.is_symlink() and dir_path.exists():
                dst = dir_path.readlink()
                boot_fs.append((dst.name, boot_vol, dst.get().top.fs))

        for uuid, name, fs in boot_fs:
            log.info("Acquiring /vmfs/volumes/%s (%s)", uuid, name)
            base = f"fs/{uuid}:{name}"
            for path in fs.path("/").rglob("*"):
                if not path.is_file():
                    continue
                collector.collect_file(path, outpath=path, base=base)


@register_module("--esxi")
class ESXi(Module):
    DESC = "ESXi interesting files"
    SPEC = [
        ("dir", "/scratch/log"),
        ("dir", "/locker/packages/var"),
        # ESXi 7
        ("dir", "/scratch/cache"),
        ("dir", "/scratch/vmkdump"),
        # ESXi 6
        ("dir", "/scratch/vmware"),
    ]


@register_module("--vmfs")
class VMFS(Module):
    DESC = "ESXi VMFS metadata files"

    @classmethod
    def _run(cls, target, collector):
        for uuid, name, fs in iter_esxi_filesystems(target):
            if not fs.__fstype__ == "vmfs":
                continue

            log.info("Acquiring /vmfs/volumes/%s (%s)", uuid, name)
            base = f"fs/{uuid}:{name}"
            for path in fs.path("/").glob("*.sf"):
                if not path.is_file():
                    continue
                collector.collect_file(path, outpath=path, base=base)


@register_module("--activities-cache")
class ActivitiesCache(Module):
    DESC = "user's activities caches"
    SPEC = [
        ("glob", "AppData/Local/ConnectedDevicesPlatform/*/ActivitiesCache.db", from_user_home),
    ]


@register_module("--hashes")
@module_arg(
    "--hash-func",
    action="append",
    type=HashFunc,
    choices=[h.value for h in HashFunc],
    help="Hash function to use",
)
@module_arg("--dir-to-hash", action="append", help="Hash only files in a provided directory")
@module_arg("--ext-to-hash", action="append", help="Hash only files with the extensions provided")
@module_arg("--glob-to-hash", action="append", help="Hash only files that match provided glob")
class FileHashes(Module):
    DESC = "file hashes"

    DEFAULT_HASH_FUNCS = (HashFunc.MD5, HashFunc.SHA1, HashFunc.SHA256)
    DEFAULT_EXTENSIONS = (
        "bat",
        "cmd",
        "com",
        "dll",
        "exe",
        "installlog",
        "installutil",
        "js",
        "lnk",
        "ps1",
        "sys",
        "tlb",
        "vbs",
    )
    DEFAULT_PATHS = ("sysvol/Windows/",)

    MAX_FILE_SIZE_BYTES = 100 * 1024 * 1024  # 100MB

    DEFAULT_FILE_FILTERS = (
        functools.partial(filter_out_by_path_match, re_pattern="^/(sysvol/)?Windows/WinSxS/"),
        functools.partial(filter_out_huge_files, max_size_bytes=MAX_FILE_SIZE_BYTES),
        functools.partial(filter_out_by_value_match, value=b"MZ", offsets=[0, 3]),
    )

    @classmethod
    def run(cls, target, cli_args, collector):
        log.info("*** Acquiring file hashes")

        specs = cls.get_specs(cli_args)

        collector.bind(cls)
        try:
            start = time.time()

            path_hashes = collect_hashes(target, specs, path_filters=cls.DEFAULT_FILE_FILTERS)
            rows_count, csv_compressed_bytes = serialize_into_csv(path_hashes, compress=True)

            collector.write_bytes(
                f"{collector.base}/{collector.METADATA_BASE}/file-hashes.csv.gz",
                csv_compressed_bytes,
            )
            log.info("Hashing is done, %s files processed in %.2f secs", rows_count, (time.time() - start))
        finally:
            collector.unbind()

    @classmethod
    def get_specs(cls, cli_args):
        path_selectors = []

        if cli_args.ext_to_hash:
            extensions = cli_args.ext_to_hash
        else:
            extensions = cls.DEFAULT_EXTENSIONS

        if cli_args.dir_to_hash or cli_args.glob_to_hash:
            if cli_args.glob_to_hash:
                path_selectors.extend([("glob", glob) for glob in cli_args.glob_to_hash])

            if cli_args.dir_to_hash:
                path_selectors.extend([("dir", (dir_path, extensions)) for dir_path in cli_args.dir_to_hash])

        else:
            path_selectors.extend([("dir", (dir_path, extensions)) for dir_path in cls.DEFAULT_PATHS])

        if cli_args.hash_func:
            hash_funcs = cli_args.hash_func
        else:
            hash_funcs = cls.DEFAULT_HASH_FUNCS

        return [(path_selector, hash_funcs) for path_selector in path_selectors]


@register_module("--handles")
@module_arg(
    "--handle-types",
    action="extend",
    help="Collect only specified handle types",
    type=NamedObjectType,
    choices=[h.value for h in NamedObjectType],
    nargs="*",
)
@local_module
class OpenHandles(Module):
    DESC = "Open handles"

    @classmethod
    def run(cls, target: Target, cli_args: dict[str, any], collector: Collector):
        if not sys.platform == "win32":
            log.error("Open Handles plugin can only run on Windows systems! Skipping...")
            return

        from acquire.dynamic.windows.collect import collect_open_handles
        from acquire.dynamic.windows.handles import serialize_handles_into_csv

        log.info("*** Acquiring open handles")

        handle_types = cli_args.handle_types

        collector.bind(cls)
        try:
            handles = collect_open_handles(handle_types)
            csv_compressed_handles = serialize_handles_into_csv(handles)

            collector.write_bytes(
                f"{collector.base}/{collector.METADATA_BASE}/open_handles.csv.gz",
                csv_compressed_handles,
            )
            log.info("Collecting open handles is done.")
        finally:
            collector.unbind()


def print_disks_overview(target):
    log.info("// Disks")
    try:
        for disk in target.disks:
            log.info("%s", disk)
            if not disk.vs:
                continue

            for volume in disk.vs.volumes:
                log.info("- %s", volume)
    except Exception:
        log.error("Failed to iterate disks")
    log.info("")


def print_volumes_overview(target):
    log.info("// Volumes")
    try:
        for volume in target.volumes:
            log.info("%s", volume)
    except Exception:
        log.error("Failed to iterate volumes")
    log.info("")


def print_acquire_warning(target: Target) -> None:
    if target.os != "windows":
        log.warning("========================================== WARNING ==========================================")
        log.warning("")
        log.warning(
            "The support for operating system '%s' is experimental. Some artifacts may not yet be included and some ",
            target.os,
        )
        log.warning("features may not work as expected. Please notify upstream for any missing artifacts or features.")
        log.warning("")
        log.warning("========================================== WARNING ==========================================")


def acquire_target(target: Target, args: argparse.Namespace, output_ts: Optional[str] = None):
    output_ts = output_ts or get_utc_now_str()
    if args.log_to_dir:
        log_file = args.log_path.joinpath(format_output_name("Unknown", output_ts, "log"))
        # This will also rename the log file on disk, which was opened in main(), if the name is different
        reconfigure_log_file(log, log_file, delay=True)
    else:
        log_file = args.log_path

    files = []
    skip_list = set()
    if log_file:
        files.append(log_file)
        if target.path.name == "local":
            skip_list.add(normalize_path(target, log_file, resolve=True))

    print_disks_overview(target)
    print_volumes_overview(target)

    if not target._os_plugin:
        log.error("Error: Unable to detect OS")
        return files

    hostname = ""
    try:
        hostname = target.hostname
    except Exception:
        log.exception("Failed to get hostname")

    version = None
    try:
        version = target.version
    except Exception:
        log.exception("Failed to detect OS version")

    if version is None:
        os_plugin_name = target._os_plugin.__name__.lower()
        version = f"{target.os} ({os_plugin_name})"

    log.info("Target name: %s", target.name)
    log.info("Hostname: %s", hostname)
    log.info("OS: %s", version)
    log.info("")

    # Prepare targets if necessary
    if target.os == "windows":
        mount_all_ntfs_filesystems(target)

    print_acquire_warning(target)

    modules_selected = {}
    modules_successful = []
    modules_failed = {}
    for name, mod in MODULES.items():
        name_slug = name.lower()
        # check if module was set in the arguments provided
        if getattr(args, name_slug):
            modules_selected[name] = mod

    profile = args.profile

    # Set profile to default if no profile, modules, files, directories or globes were selected
    if not profile and not modules_selected and not args.file and not args.directory and not args.glob:
        log.info("Using default collection profile")
        profile = "default"
        log.info("")

    if profile and profile != "none":
        if target.os not in PROFILES[profile]:
            log.error("No collection set for OS %s with profile %s", target.os, profile)
            return files

        for mod in PROFILES[profile][target.os]:
            modules_selected[mod.__modname__] = mod

    log.info("Modules selected: %s", ", ".join(sorted(modules_selected)))

    local_only_modules = {name: module for name, module in modules_selected.items() if hasattr(module, "__local__")}
    if target.path.name != "local" and local_only_modules:
        for name, module in local_only_modules.items():
            modules_failed[module.__name__] = "Not running on a local target"
        log.error(
            "Can not use local-only modules with non-local targets. Skipping: %s",
            " ".join(sorted(local_only_modules.keys())),
        )
        log.info("")
        # Remove local-only modules from the modules list
        modules_selected = dict(modules_selected.items() - local_only_modules.items())

    log_file_handler = get_file_handler(log)
    # Prepare log file and output file names
    if log_file_handler and args.log_to_dir:
        log_file = format_output_name(target.name, output_ts, "log")
        # This will also rename the log file on disk, which was opened and written previously.
        log_file_handler.set_filename(log_file)
        log_path = Path(log_file_handler.baseFilename).resolve()
        log.info("Logging to file %s", log_path)
        files = [log_file_handler.baseFilename]
        if target.path.name == "local":
            skip_list = {normalize_path(target, log_path, resolve=True)}

    output_path = args.output
    if output_path.is_dir():
        output_dir = format_output_name(target.name, output_ts)
        output_path = output_path.joinpath(output_dir)
    output_path = output_path.resolve()

    output = OUTPUTS[args.output_type](
        output_path,
        compress=args.compress,
        encrypt=args.encrypt,
        public_key=args.public_key,
    )
    files.append(output.path)
    if target.path.name == "local":
        skip_list.add(normalize_path(target, output.path, resolve=True))

    log.info("Writing output to %s", output.path)
    if skip_list:
        log.info("Skipping own files: %s", ", ".join(skip_list))
    log.info("")

    dir_base = "fs"
    if target.os != "windows":
        dir_base = "fs/$rootfs$"

    with Collector(target, output, base=dir_base, skip_list=skip_list) as collector:
        # Acquire specified files
        if args.file or args.directory or args.glob:
            log.info("*** Acquiring specified paths")
            spec = []

            if args.file:
                for path in args.file:
                    spec.append(("file", path.strip()))

            if args.directory:
                for path in args.directory:
                    spec.append(("dir", path.strip()))

            if args.glob:
                for path in args.glob:
                    spec.append(("glob", path.strip()))

            collector.collect(spec, module_name=CLI_ARGS_MODULE)
            modules_successful.append(CLI_ARGS_MODULE)
            log.info("")

        # Run modules (sort first based on execution order)
        modules_selected = sorted(modules_selected.items(), key=lambda module: module[1].EXEC_ORDER)
        for name, mod in modules_selected:
            try:
                mod.run(target, args, collector)

                modules_successful.append(mod.__name__)
            except Exception:
                log.error("Error while running module %s", name, exc_info=True)
                modules_failed[mod.__name__] = get_formatted_exception()
            log.info("")

        collection_report = collector.report

    log.info("Done collecting artifacts:")

    # prepare and render full report only if logging level is more permissive than INFO
    if log.level < logging.INFO:
        log.debug(get_full_formatted_report(collection_report))

    log.info(get_report_summary(collection_report))

    if not args.disable_report:
        collection_report_serialized = collection_report.get_records_per_module_per_outcome(serialize_records=True)

        execution_report = {
            "target": str(target),
            "name": target.name,
            "timestamp": get_utc_now().isoformat(),
            "modules-successful": modules_successful,
            "modules-failed": modules_failed,
            **collection_report_serialized,
        }

        if args.output.is_dir():
            report_file_name = format_output_name(target.name, postfix=output_ts, ext="report.json")
        else:
            report_file_name = f"{output_path.name}.report.json"

        report_file_path = output_path.parent / report_file_name
        persist_execution_report(report_file_path, execution_report)

        files.append(report_file_path)
        log.info("Acquisition report for %s is written to %s", target, report_file_path)

    log.info("Output: %s", output.path)
    return files


def upload_files(
    paths: list[Path],
    upload_plugin: UploaderPlugin,
    no_proxy: bool = False,
):
    proxies = None if no_proxy else urllib.request.getproxies()
    log.debug("Proxies: %s (no_proxy = %s)", proxies, no_proxy)

    try:
        upload_files_using_uploader(upload_plugin, paths, proxies)
    except Exception:
        log.error("Upload %s FAILED. See log file for details.", paths)
        log.exception("")


PROFILES = {
    "full": {
        "windows": [
            NTFS,
            EventLogs,
            Registry,
            Tasks,
            ETL,
            Recents,
            RecycleBin,
            Drivers,
            PowerShell,
            Prefetch,
            Appcompat,
            PCA,
            Syscache,
            WBEM,
            AV,
            ActivitiesCache,
            BITS,
            DHCP,
            DNS,
            History,
            Misc,
            NTDS,
            QuarantinedFiles,
            RemoteAccess,
            WindowsNotifications,
        ],
        "linux": [
            Etc,
            Boot,
            Home,
            Var,
        ],
        "bsd": [
            Etc,
            Boot,
            Home,
            Var,
            BSD,
        ],
        "esxi": [
            Bootbanks,
            ESXi,
            VMFS,
        ],
        "osx": [
            Etc,
            Home,
            Var,
            OSX,
            History,
        ],
    },
    "default": {
        "windows": [
            NTFS,
            EventLogs,
            Registry,
            Tasks,
            ETL,
            Recents,
            RecycleBin,
            Drivers,
            PowerShell,
            Prefetch,
            Appcompat,
            PCA,
            Syscache,
            WBEM,
            AV,
            BITS,
            DHCP,
            DNS,
            Misc,
            RemoteAccess,
            ActivitiesCache,
        ],
        "linux": [
            Etc,
            Boot,
            Home,
            Var,
        ],
        "bsd": [
            Etc,
            Boot,
            Home,
            Var,
            BSD,
        ],
        "esxi": [
            Bootbanks,
            ESXi,
            VMFS,
        ],
        "osx": [
            Etc,
            Home,
            Var,
            OSX,
        ],
    },
    "minimal": {
        "windows": [
            NTFS,
            EventLogs,
            Registry,
            Tasks,
            PowerShell,
            Prefetch,
            Appcompat,
            PCA,
            Misc,
        ],
        "linux": [
            Etc,
            Boot,
            Home,
            Var,
        ],
        "bsd": [
            Etc,
            Boot,
            Home,
            Var,
            BSD,
        ],
        "esxi": [
            Bootbanks,
            ESXi,
        ],
        "osx": [
            Etc,
            Home,
            Var,
            OSX,
        ],
    },
    "none": None,
}


def main():
    parser = create_argument_parser(PROFILES, MODULES)
    args = parse_acquire_args(parser, config=CONFIG)

    try:
        check_and_set_log_args(args)
    except ValueError as err:
        parser.exit(err)

    if args.log_to_dir:
        # When args.upload files are specified, only these files are uploaded
        # and no other action is done. Thus a log file specifically named
        # Upload_<date>.log is created
        file_prefix = "Upload" if args.upload else "Unknown"
        log_file = args.log_path.joinpath(format_output_name(file_prefix, args.start_time, "log"))
    else:
        log_file = args.log_path

    setup_logging(log, log_file, args.verbose, delay=args.log_delay)

    log.info(ACQUIRE_BANNER)
    log.info("User: %s | Admin: %s", get_user_name(), is_user_admin())
    log.info("Arguments: %s", " ".join(sys.argv[1:]))
    log.info("Default Arguments: %s", " ".join(args.config.get("arguments")))
    log.info("")

    plugins_to_load = [("cloud", MinIO)]
    upload_plugins = UploaderRegistry("acquire.plugins", plugins_to_load)

    try:
        check_and_set_acquire_args(args, upload_plugins)
    except ValueError as err:
        log.exception(err)
        parser.exit(1)

    if args.upload:
        try:
            upload_files(args.upload, args.upload_plugin, args.no_proxy)
        except Exception:
            log.exception("Failed to upload files")
        return

    RemoteStreamConnection.configure(args.cagent_key, args.cagent_certificate)

    target_path = args.target

    if target_path == "local":
        target_query = {}
        if args.force_fallback:
            target_query.update({"force-directory-fs": 1})

        if args.fallback:
            target_query.update({"fallback-to-directory-fs": 1})

        target_query = urllib.parse.urlencode(target_query)
        target_path = f"{target_path}?{target_query}"

    log.info("Loading target %s", target_path)

    try:
        target = Target.open(target_path)
        log.info(target)
    except Exception:
        if not is_user_admin():
            log.error("Failed to load target, try re-running as administrator/root.")
            parser.exit(1)
        log.exception("Failed to load target")
        raise

    if target.os == "esxi":
        # Loader found that we are running on an esxi host
        # Perform operations to "enhance" memory
        with esxi_memory_context_manager():
            acquire_children_and_targets(target, args)
    else:
        acquire_children_and_targets(target, args)


def load_child(target: Target, child_path: Path) -> None:
    log.info("")
    log.info("Loading child target %s", child_path)
    try:
        child = target.open_child(child_path)
        log.info(target)
    except Exception:
        log.exception("Failed to load child target")
        raise

    return child


def acquire_children_and_targets(target: Target, args: argparse.Namespace):
    if args.child:
        target = load_child(target, args.child)

    log.info("")

    try:
        files = acquire_target(target, args, args.start_time)
    except Exception:
        log.exception("Failed to acquire target")
        raise

    if args.children:
        for child in target.list_children():
            try:
                child_target = load_child(target, child.path)
            except Exception:
                continue

            log.info("")

            try:
                child_files = acquire_target(child_target, args)
                files.extend(child_files)
            except Exception:
                log.exception("Failed to acquire child target")
                continue

    files = sort_files(files)

    if args.auto_upload:
        log_file_handler = get_file_handler(log)
        if log_file_handler:
            log_file_handler.close()

        log.info("")
        try:
            upload_files(files, args.upload_plugin)
        except Exception:
            log.exception("Failed to upload files")


def sort_files(files: list[Union[str, Path]]) -> list[Path]:
    log_files: list[Path] = []
    tar_paths: list[Path] = []
    report_paths: list[Path] = []

    suffix_map = {".log": log_files, ".json": report_paths}

    for file in files:
        if isinstance(file, str):
            file = Path(file)

        suffix_map.get(file.suffix, tar_paths).append(file)

    # Reverse log paths, as the first one in ``files`` is the main one.
    log_files.reverse()

    return tar_paths + report_paths + log_files


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
    except Exception:
        sys.exit(1)



from __future__ import annotations

import ctypes
from ctypes.wintypes import (
    BOOL,
    DWORD,
    HANDLE,
    LPDWORD,
    LPVOID,
    LPWSTR,
    PHANDLE,
    PULONG,
    ULONG,
    USHORT,
    WCHAR,
)
from enum import IntEnum

PVOID = ctypes.c_void_p
NTSTATUS = ULONG
NULL = None

ULONG_PTR = ctypes.c_size_t


class ProcessToken(IntEnum):
    TOKEN_QUERY = 0x0008
    TOKEN_ADJUST_PRIVILEGES = 0x0020


class ProcessAccess(IntEnum):
    PROCESS_TERMINATE = 0x0001
    PROCESS_CREATE_THREAD = 0x0002
    PROCESS_VM_OPERATION = 0x0008
    PROCESS_VM_READ = 0x0010
    PROCESS_VM_WRITE = 0x0020
    PROCESS_DUP_HANDLE = 0x0040
    PROCESS_CREATE_PROCESS = 0x0080
    PROCESS_SET_QUOTA = 0x0100
    PROCESS_SET_INFORMATION = 0x0200
    PROCESS_QUERY_INFORMATION = 0x0400
    SYNCHRONIZE = 0x00100000
    PROCESS_ALL_ACCESS = 0x1F0FFF


class ErrorCode(IntEnum):
    ERROR_SUCCESS = 0x0
    ERROR_ACCESS_DENIED = 0x5
    ERROR_INVALID_PARAMETER = 0x57
    ERROR_PARTIAL_COPY = 0x12B
    ERROR_NOT_ALL_ASSIGNED = 0x514


class DuplicateHandleFlags(IntEnum):
    DUPLICATE_CLOSE_SOURCE = 0x00000001
    DUPLICATE_SAME_ACCESS = 0x00000002
    DUPLICATE_SAME_ATTRIBUTES = 0x00000004


class SYSTEM_INFORMATION_CLASS(IntEnum):
    SystemHandleInformation = 0x10
    SystemExtendedHandleInformation = 0x40


class OBJECT_INFORMATION_CLASS(IntEnum):
    ObjectBasicInformation = 0
    ObjectNameInformation = 1
    ObjectTypeInformation = 2


class FILE_INFORMATION_CLASS(IntEnum):
    FileNameInformation = 9


class SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX(ctypes.Structure):
    _fields_ = [
        ("Object", PVOID),
        ("UniqueProcessId", ULONG_PTR),
        ("HandleValue", ULONG_PTR),
        ("GrantedAccess", ULONG),
        ("CreatorBackTraceIndex", USHORT),
        ("ObjectTypeIndex", USHORT),
        ("HandleAttributes", ULONG),
        ("Reserved", ULONG),
    ]

    @property
    def object(self) -> str:
        return hex(self.Object)

    @property
    def unique_process_id(self) -> str:
        return str(self.UniqueProcessId)

    @property
    def handle_value(self) -> str:
        return str(self.HandleValue)

    @property
    def granted_access(self) -> str:
        return str(self.GrantedAccess)

    @property
    def creator_back_trace_index(self) -> str:
        return str(self.CreatorBackTraceIndex)

    @property
    def object_type_index(self) -> str:
        return str(self.ObjectTypeIndex)

    @property
    def handle_attributes(self) -> str:
        return str(self.HandleAttributes)

    @property
    def reserved(self) -> str:
        return str(self.Reserved)


class SYSTEM_HANDLE_INFORMATION_EX(ctypes.Structure):
    _fields_ = [
        ("NumberOfHandles", ULONG_PTR),
        ("Reserved", ULONG_PTR),
        ("Handles", SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX * 1),
    ]


def FileNameInformationFactory(file_name_size: int = 1):
    class FILE_NAME_INFORMATION(ctypes.Structure):
        _fields_ = [("FileNameLength", ULONG), ("FileName", WCHAR * file_name_size)]

    return FILE_NAME_INFORMATION()


class IO_STATUS_BLOCK_DUMMYUNIONNAME(ctypes.Union):
    _fields_ = [("Status", NTSTATUS), ("Pointer", ULONG_PTR)]


class IO_STATUS_BLOCK(ctypes.Structure):
    _fields_ = [("DUMMYUNIONNAME", IO_STATUS_BLOCK_DUMMYUNIONNAME), ("Information", ctypes.c_size_t)]


class LUID(ctypes.Structure):
    _fields_ = [
        ("LowPart", DWORD),
        ("HighPart", DWORD),
    ]


class LUID_AND_ATTRIBUTES(ctypes.Structure):
    _fields_ = [
        ("Luid", LUID),
        ("Attributes", DWORD),
    ]


class TOKEN_PRIVILEGES(ctypes.Structure):
    _fields_ = [
        ("PrivilegeCount", DWORD),
        ("Privileges", 1 * LUID_AND_ATTRIBUTES),
    ]


class Handle:
    """Handle object"""

    def __init__(self, handle: SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX, handle_type: str, handle_name: str) -> None:
        self.name = handle_name
        self.handle_type = handle_type

        self.object = handle.object
        self.unique_process_id = handle.unique_process_id
        self.handle_value = handle.handle_value
        self.granted_access = handle.granted_access
        self.creator_back_trace_index = handle.creator_back_trace_index
        self.object_type_index = handle.object_type_index
        self.handle_attributes = handle.handle_attributes
        self.reserved = handle.reserved
        self._handle = handle

    @property
    def dictionary(self):
        return {key: value for key, value in self.__dict__.items() if not key.startswith("_")}


class UNICODE_STRING(ctypes.Structure):
    _fields_ = [
        ("Length", USHORT),
        ("MaximumLength", USHORT),
        ("Buffer", LPWSTR),
    ]

    def __str__(self) -> str:
        return self.Buffer

    @classmethod
    def from_str(cls, value: str) -> UNICODE_STRING:
        """Initializes a UNICODE_STRING structure."""
        destination = cls()
        value_buffer = ctypes.create_unicode_buffer(value)

        ctypes.memset(ctypes.addressof(destination), 0, ctypes.sizeof(destination))
        destination.Buffer = ctypes.cast(value_buffer, LPWSTR)
        destination.Length = ctypes.sizeof(value_buffer) - 2  # Excluding terminating NULL character
        destination.MaximumLength = destination.Length

        return destination


class PUBLIC_OBJECT_TYPE_INFORMATION(ctypes.Structure):
    _fields_ = [
        ("Name", UNICODE_STRING),
        ("Reserved", ULONG * 22),
    ]

    @property
    def name(self) -> str:
        return str(self.Name)


PUNICODE_STRING = ctypes.POINTER(UNICODE_STRING)


class OBJECT_DIRECTORY_INFORMATION(ctypes.Structure):
    _fields_ = [
        ("Name", UNICODE_STRING),
        ("TypeName", UNICODE_STRING),
    ]

    @property
    def name(self) -> str:
        return str(self.Name)

    @property
    def type_name(self) -> str:
        return str(self.TypeName)


__all__ = [
    "BOOL",
    "DWORD",
    "HANDLE",
    "LPWSTR",
    "PHANDLE",
    "PULONG",
    "ULONG",
    "USHORT",
    "PVOID",
    "NTSTATUS",
    "NULL",
    "UNICODE_STRING",
    "PUNICODE_STRING",
    "OBJECT_DIRECTORY_INFORMATION",
    "ProcessToken",
    "ProcessAccess",
    "ErrorCode",
    "DuplicateHandleFlags",
    "SYSTEM_INFORMATION_CLASS",
    "OBJECT_INFORMATION_CLASS",
    "FILE_INFORMATION_CLASS",
    "SYSTEM_HANDLE_TABLE_ENTRY_INFO_EX",
    "SYSTEM_HANDLE_INFORMATION_EX",
    "PUBLIC_OBJECT_TYPE_INFORMATION",
    "IO_STATUS_BLOCK_DUMMYUNIONNAME",
    "IO_STATUS_BLOCK",
    "LUID",
    "LUID_AND_ATTRIBUTES",
    "TOKEN_PRIVILEGES",
    "Handle",
    "WCHAR",
    "LPVOID",
    "LPDWORD",
]



import csv
import gzip
import hashlib
import io
import logging
import re
import time
from pathlib import Path
from typing import Callable, Generator, Iterable, List, Optional, Set, Tuple

from dissect.target import Target
from dissect.target.helpers.fsutil import TargetPath

from acquire.utils import StrEnum

log = logging.getLogger(__name__)


class HashFunc(StrEnum):
    MD5 = "md5"
    SHA1 = "sha1"
    SHA256 = "sha256"

    def as_hashlib_method(self):
        if self == HashFunc.MD5:
            return hashlib.md5
        elif self == HashFunc.SHA1:
            return hashlib.sha1
        elif self == HashFunc.SHA256:
            return hashlib.sha256


PROGRESS_LOG_STEP = 10000  # processed files

CSV_COLUMNS = ["path", "file-size"] + [h.value for h in HashFunc]


def get_paths_from_dir(
    target: Target,
    glob: str,
    extensions: Optional[Set[str]] = None,
) -> Generator[Path, None, None]:
    """Yield paths that match provided `glob` pattern and `extensions` values"""

    extension_suffixes = {f".{ext}" for ext in extensions} if extensions else None

    for path in target.fs.path("/").glob(glob):
        if extensions and path.suffix not in extension_suffixes:
            continue
        yield path


def get_path_details(path: TargetPath, hash_funcs: Optional[Iterable[HashFunc]] = None) -> Tuple:
    """
    Calculate and return the details for specified path.

    The details include file size and hashes, calculated for
    hash functions provided in `hash_funcs`.
    """

    hash_funcs = hash_funcs or []
    if hash_funcs:
        provided_hash_funcs_sorted = sorted(set(hash_funcs))

        hashes = path.get().hash(algos=[f.as_hashlib_method() for f in provided_hash_funcs_sorted])

        hashes_map = dict(zip(provided_hash_funcs_sorted, hashes))
    else:
        hashes_map = {}

    file_size = path.stat().st_size

    return {
        "path": str(path),
        "file-size": file_size,
        **{func.value: digest for func, digest in hashes_map.items()},
    }


def filter_out_nonfiles(paths: Iterable[Path]) -> Generator[Path, None, None]:
    """Filter out paths that are not files"""
    for path in paths:
        try:
            is_file = path.is_file()
        except OSError:
            log.debug("Exception while determining file properties of `%s`, skipping", path, exc_info=True)
            continue

        if not is_file:
            continue

        yield path


def filter_out_huge_files(paths: Iterable[Path], *, max_size_bytes: int) -> Iterable[Path]:
    """Filter out paths that are larger than `max_size_bytes` value"""
    for path in paths:
        try:
            file_size = path.stat().st_size
        except Exception:
            log.debug("Exception while getting the size of `%s`, skipping", path, exc_info=True)
            continue

        if file_size > max_size_bytes:
            continue

        yield path


def filter_out_by_value_match(
    paths: Iterable[Path],
    *,
    value: bytes,
    offsets: Iterable[int] = (0,),
) -> Generator[Path, None, None]:
    """Filter out paths where file data matches the provided `value` at the specified offsets"""

    if not offsets:
        raise ValueError("No offsets provided")

    value_len = len(value)
    bytes_to_read = value_len + max(offsets)

    for path in paths:
        try:
            fh = path.open("rb")
        except Exception:
            log.debug("Exception while opening path `%s`, skipping", path, exc_info=True)
            continue

        buffer = fh.read(bytes_to_read)
        for offset in offsets:
            if buffer[offset : offset + value_len] == value:
                continue

        yield path


def filter_out_by_path_match(
    paths: Iterable[Path],
    *,
    re_pattern: str,
    re_flags: re.RegexFlag = re.IGNORECASE,
) -> Iterable[Path]:
    """Filter out paths that match provided regex pattern"""
    pattern = re.compile(re_pattern, flags=re_flags)
    return filter(lambda p: not pattern.match(str(p)), paths)


def collect_hashes(
    target: Target,
    specs: Iterable[Iterable[Tuple]],
    path_filters: Iterable[Callable[[Iterable[Path]], Iterable[Path]]] = None,
) -> Generator[Tuple, None, None]:
    """
    Walk through the paths, calculate hashes and return details per path.

    Spec contains a path selector and a list of hash functions to compute against the paths.
    For example:
        [
            ("dir", ("sysvol/Windows/", ("exe", "dll", "sys"))),
            (HashFunc.MD5, HashFunc.SHA1)
        ]
    """

    log.info("Starting to collect hashes for spec: %s", specs)

    stream_hash_func_pairs = []

    for spec in filter(lambda spec: spec[0][0] == "glob", specs):
        path_selector, hash_funcs = spec
        glob_value = path_selector[1]
        paths_stream = target.fs.path("/").glob(glob_value)

        stream_hash_func_pairs.append((paths_stream, hash_funcs))

    for spec in filter(lambda spec: spec[0][0] == "dir", specs):
        path_selector, hash_funcs = spec
        path_selector_param = path_selector[1]

        if isinstance(path_selector_param, (tuple, list)):
            dir_name, extensions = path_selector_param
            extensions = set(extensions) if extensions else set()
        else:
            dir_name = path_selector_param
            extensions = None

        # Create a glob to recursively catch all files inside a directory
        dir_glob = str(target.fs.path(dir_name).joinpath("**/*"))

        paths_stream = get_paths_from_dir(target, dir_glob, extensions)
        stream_hash_func_pairs.append((paths_stream, hash_funcs))

    path_filters = path_filters or []

    for paths, hash_funcs in stream_hash_func_pairs:
        seen_paths = set()

        paths = filter_out_nonfiles(paths)

        for filter_func in path_filters:
            paths = filter_func(paths)

        for path in paths:
            path_str = str(path)
            if path_str in seen_paths:
                continue

            seen_paths.add(path_str)

            try:
                details = get_path_details(path, hash_funcs)
            except Exception:
                log.debug("Error while processing path `{path}`, skipping", exc_info=True)
                continue

            yield details


def serialize_into_csv(rows: Iterable[List], compress: bool = True) -> Tuple[int, bytes]:
    """
    Serialize provided rows into normal or gzip-compressed CSV, and return a tuple
    containing the number of rows processed and the result bytes.
    """

    raw_buffer = io.BytesIO()

    counter = 0
    start = time.time()

    if compress:
        buffer = gzip.GzipFile(fileobj=raw_buffer, mode="wb")
    else:
        buffer = raw_buffer

    with io.TextIOWrapper(buffer, encoding="utf-8") as wrapper:
        csv_writer = csv.DictWriter(wrapper, fieldnames=CSV_COLUMNS)
        csv_writer.writeheader()
        for row in rows:
            csv_writer.writerow(row)
            counter += 1
            if counter % PROGRESS_LOG_STEP == 0:
                log.info(
                    "%s files processed in %.2f secs, last row: %s",
                    counter,
                    (time.time() - start),
                    row,
                )

    return (counter, raw_buffer.getvalue())



import argparse
import ctypes
import datetime
import getpass
import json
import os
import pathlib
import re
import sys
import textwrap
import traceback
from enum import Enum
from io import SEEK_SET, UnsupportedOperation
from pathlib import Path
from stat import S_IRGRP, S_IROTH, S_IRUSR
from typing import Any, Optional

from dissect.target import Target
from dissect.util.stream import AlignedStream

from acquire.outputs import OUTPUTS
from acquire.uploaders.plugin_registry import UploaderRegistry

try:
    # Windows systems do not have the fcntl module.
    from fcntl import F_SETFL, fcntl

    HAS_FCNTL = True
except ImportError:
    HAS_FCNTL = False


class VolatileStream(AlignedStream):
    """Streaming class to handle various procfs and sysfs edge-cases.  Backed by `AlignedStream`.

    Args:
        path: Path of the file to obtain a file-handle from.
        mode: Mode string to open the file-handle with. Such as "rt" and "rb".
        flags: Flags to open the file-descriptor with.
        size: The maximum size of the stream. None if unknown.
    """

    def __init__(
        self,
        path: Path,
        mode: str = "rb",
        # Windows and Darwin systems don't have O_NOATIME or O_NONBLOCK. Add them if they are available.
        flags: int = (os.O_RDONLY | getattr(os, "O_NOATIME", 0) | getattr(os, "O_NONBLOCK", 0)),
        size: int = 1024 * 1024 * 5,
    ):
        self.fh = path.open(mode)
        self.fd = self.fh.fileno()

        if HAS_FCNTL:
            fcntl(self.fd, F_SETFL, flags)

        st_mode = os.fstat(self.fd).st_mode
        write_only = (st_mode & (S_IRUSR | S_IRGRP | S_IROTH)) == 0  # novermin

        super().__init__(0 if write_only else size)

    def seek(self, pos: int, whence: int = SEEK_SET) -> int:
        raise UnsupportedOperation("VolatileStream is not seekable")

    def seekable(self) -> bool:
        return False

    def _read(self, offset: int, length: int) -> bytes:
        result = []
        while length:
            try:
                buf = os.read(self.fd, min(length, self.size - offset))
            except BlockingIOError:
                break

            if not buf:
                break

            result.append(buf)
            offset += len(buf)
            length -= len(buf)
        return b"".join(result)


class StrEnum(str, Enum):
    """Sortable and serializible string-based enum"""


def create_argument_parser(profiles: dict, modules: dict) -> argparse.ArgumentParser:
    desc = ""

    profile_names = (name for name in profiles.keys() if name != "none")

    for name in profile_names:
        desc += f"{name} profile:\n"
        minindent = max([len(os_) for os_ in profiles[name].keys()])
        descfmt = f"  {{:{minindent}s}}: {{}}\n"
        for os_ in profiles[name].keys():
            indent = 4 + len(os_)
            modlist = textwrap.wrap(", ".join([mod.__modname__ for mod in profiles[name][os_]]), 50)

            moddesc = modlist.pop(0)
            for ml in modlist:
                moddesc += "\n" + (" " * indent) + ml

            desc += descfmt.format(os_, moddesc)
        desc += "\n"

    parser = argparse.ArgumentParser(
        prog="acquire",
        description=desc,
        epilog=(
            "If no target is specified, 'local' is used.\n\n"
            "If no options are given, the collection profile 'default' is used."
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        fromfile_prefix_chars="@",
    )

    parser.add_argument(
        "target",
        metavar="TARGET",
        default="local",
        nargs="?",
        help="target to load (default: local)",
    )
    parser.add_argument("-o", "--output", default=Path("."), type=Path, help="output directory")
    parser.add_argument(
        "-ot",
        "--output-type",
        choices=OUTPUTS.keys(),
        default="tar",
        help="output type (default: tar)",
    )
    parser.add_argument(
        "--compress",
        action="store_true",
        help="compress output (if supported by the output type)",
    )
    parser.add_argument(
        "--encrypt",
        action="store_true",
        help="encrypt output (if supported by the output type)",
    )
    parser.add_argument("--public-key", type=Path, help=argparse.SUPPRESS)
    parser.add_argument("-l", "--log", type=Path, help="log directory location")
    parser.add_argument("--no-log", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("-p", "--profile", choices=profiles.keys(), help="collection profile")

    parser.add_argument("-f", "--file", action="append", help="acquire file")
    parser.add_argument("-d", "--directory", action="append", help="acquire directory recursively")
    parser.add_argument("-g", "--glob", action="append", help="acquire files matching glob pattern")

    parser.add_argument("--disable-report", action="store_true", help="disable acquisition report file")

    parser.add_argument("--child", help="only collect specific child")
    parser.add_argument(
        "--children",
        action="store_true",
        help="collect all children in addition to main target",
    )

    parser.add_argument(
        "--force-fallback",
        action="store_true",
        help="force filesystem access directly through OS level. Only supported with target 'local'",
    )
    parser.add_argument(
        "--fallback",
        action="store_true",
        help=(
            "fallback to OS level filesystem access if filesystem type is not supported. "
            "Only supported with target 'local'"
        ),
    )

    parser.add_argument(
        "-u",
        "--auto-upload",
        action="store_true",
        help="upload result files after collection",
    )
    parser.add_argument(
        "--upload",
        nargs="+",
        help="upload specified files (all other acquire actions are ignored)",
    )
    parser.add_argument("--no-proxy", action="store_true", help="don't autodetect proxies")

    for module_cls in modules.values():
        for args, kwargs in module_cls.__cli_args__:
            parser.add_argument(*args, **kwargs)

    parser.add_argument("-v", "--verbose", action="count", default=3, help="increase output verbosity")
    return parser


def parse_acquire_args(
    parser: argparse.ArgumentParser,
    config: dict[str, Any],
) -> argparse.Namespace:
    """Parse and set the acquire command line arguments.

    The arguments are set to values supplied in ``config[arguments]``, when not
    changed from the default values specified in ``parser``.

    The ``config`` dict is added to the parsed command line arguments for
    convenience of later use.

    Args:
        parser: A parser for acquire command line arguments.
        config: A dict of global configuration values.

    Returns:
        A command line arguments namespace
    """
    command_line_args = parser.parse_args()
    _merge_args_and_config(parser, command_line_args, config)

    return command_line_args


def _merge_args_and_config(
    parser: argparse.ArgumentParser,
    command_line_args: argparse.Namespace,
    config: dict[str, Any],
):
    """Update the parsed command line arguments with the optional set of configured default arguments.

    The arguments are set to values supplied in ``config[arguments]``, when not
    changed from the default values specified in ``parser``.

    The ``config`` dict is added to the parsed command line arguments for
    convenience of later use.

    Args:
        parser: A parser for acquire command line arguments.
        command_line_args: The namespace with parsed commandline args. This
                           namespace is updated with the optional set of
                           configured default arguments.
        config: A dict of global configuration values.
    """
    config_defaults = config.get("arguments")
    if not config_defaults:
        config_defaults = config["arguments"] = []

    config_defaults_args = parser.parse_args(config_defaults)

    for argument, value in command_line_args._get_kwargs():
        if parser.get_default(argument) == value:
            config_argument = getattr(config_defaults_args, argument, value)
            setattr(command_line_args, argument, config_argument)

    setattr(command_line_args, "config", config)


def check_and_set_log_args(args: argparse.Namespace):
    """Check command line arguments which are related to logging.

    Also some arguments derived from the user supplied ones are set in the
    ``args`` namespace for convenience.

    This function is separate from ``check_and_set_acquire_args()`` as logging
    needs to be setup as soon as possible when running acquire.

    Args:
        args: The namespace containing the command line arguments.

    Raises:
        ValueError: When an invalid combination of arguments is found.
    """
    start_time = get_utc_now_str()
    log_path = None
    log_to_dir = False
    log_delay = False

    if not args.no_log:
        log_path = args.log or args.output

        if log_path.is_dir():
            log_to_dir = True
            log_delay = True
        elif log_path.is_file() or (not log_path.exists() and log_path.parent.is_dir()):
            # Logging to a single file is allowed, even if the file does not yet
            # exist, as it will be automatically created. However then the parent
            # directory must exist.
            if args.children:
                # If children are acquired, logging can only happen to separate
                # files, so log_path needs to be a directory.
                raise ValueError("Log path must be a directory when using --children")
        else:
            raise ValueError(f"Log path doesn't exist: {log_path}")

    setattr(args, "start_time", start_time)
    setattr(args, "log_path", log_path)
    setattr(args, "log_to_dir", log_to_dir)
    setattr(args, "log_delay", log_delay)


def check_and_set_acquire_args(
    args: argparse.Namespace,
    upload_plugins: UploaderRegistry,
):
    """Check the command line arguments and set some derived arguments.

    This function is separate from ``check_and_set_log_args()`` as logging
    needs to be setup as soon as possible when running acquire.

    Args:
        args: The namespace containing the command line arguments.
        upload_plugins: The registry of available upload plugins.

    Raises:
        ValueError: When an invalid combination of arguments is found.
    """
    upload_plugin = None

    # check & set upload related configuration
    if args.upload or args.auto_upload:
        upload_mode = args.config.get("upload", {}).get("mode")
        if not upload_mode:
            raise ValueError("Uploading is not configured")

        upload_plugin_cls = upload_plugins.get(upload_mode)
        if not upload_plugin_cls:
            raise ValueError(f"Invalid upload mode: {upload_mode}")

        # If initialization of the plugin fails, a ValueError is raised
        upload_plugin = upload_plugin_cls(**args.config)

    setattr(args, "upload_plugin", upload_plugin)

    if not args.upload:
        # check output related configuration
        if args.children and not args.output.is_dir():
            raise ValueError("Output path must be a directory when using --children")
        elif not args.output.exists() and not args.output.parent.is_dir():
            raise ValueError(f"Output path doesn't exist: {args.output}")

        # check & set encryption related configuration
        if args.encrypt:
            public_key = args.config.get("public_key")
            if not public_key and args.public_key and args.public_key.is_file():
                public_key = args.public_key.read_text()
            if not public_key:
                raise ValueError("No public key available (embedded or argument)")
            setattr(args, "public_key", public_key)

        # set cagent related configuration
        setattr(args, "cagent_key", args.config.get("cagent_key"))
        setattr(args, "cagent_certificate", args.config.get("cagent_certificate"))


def get_user_name() -> str:
    try:
        return getpass.getuser()
    except Exception:
        return "unknown"


def is_user_admin() -> bool:
    try:
        return os.geteuid() == 0
    except AttributeError:
        pass

    try:
        return ctypes.windll.shell32.IsUserAnAdmin() == 1
    except Exception:
        return False


def get_utc_now() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


def get_utc_now_str() -> str:
    return get_utc_now().strftime("%Y%m%d%H%M%S")


def get_formatted_exception() -> str:
    exc_info = sys.exc_info()
    return "".join(traceback.format_exception(*exc_info))


def format_output_name(prefix: str, postfix: Optional[str] = None, ext: Optional[str] = None) -> str:
    if not postfix:
        postfix = datetime.datetime.utcnow().strftime("%Y%m%d%H%M%S")
    name = f"{prefix}_{postfix}"
    if ext:
        name = f"{name}.{ext}"
    return name


def persist_execution_report(path: Path, report_data: dict) -> Path:
    with open(path, "w") as f:
        f.write(json.dumps(report_data, sort_keys=True, indent=4))


SYSVOL_SUBST = re.compile(r"^(/\?\?/)?[cC]:")


def normalize_path(target: Target, path: pathlib.Path, resolve: bool = False) -> str:
    if resolve:
        path = path.resolve()

    path = path.as_posix()

    if not target.fs.case_sensitive:
        path = path.lower()

    if target.os == "windows":
        # As dissect.target always maps c: onto sysvol, we can do this substitution here.
        path = SYSVOL_SUBST.sub("sysvol", path)

    return path



import logging
import re
import subprocess
from contextlib import contextmanager


@contextmanager
def esxi_memory_context_manager():
    memory = EsxiMemoryManager()

    try:
        memory.setup()
        yield memory
    finally:
        memory.reset()


class EsxiMemoryManager:
    def __init__(self) -> None:
        self.group_id: str = ""
        self.mem_scheme: dict[str, str] = dict()

    def setup(self) -> None:
        """Change the ESXi memory limits of the group to unlimited."""
        self.group_id = self._get_group_id()
        self.mem_scheme = self._get_memory_scheme()
        self._set_memory_limits(max="unlimited", min_limit="unlimited")

    def reset(self) -> None:
        """Put memory limits back to their original values."""
        if self.mem_scheme:
            self._set_memory_limits(max=self.mem_scheme.get("max"), min_limit=self.mem_scheme.get("minLimit"))

    def _execute_vsish_command(self, command: list[str]) -> str:
        """Performs and logs a vsish command."""
        vsish_command = ["vsish", "-e"] + command
        logging.info(f"Executing '{' '.join(vsish_command)}' on ESXi host.")

        output = subprocess.check_output(vsish_command)

        return output.decode("utf-8")

    def _get_group_id(self) -> str:
        """Get the group ID of the current session."""
        group_id = self._execute_vsish_command(
            [
                "set",
                "/sched/groupPathNameToID",
                "host",
                "vim",
                "vimuser",
                "terminal",
                "ssh",
            ]
        )
        group_id = group_id.split(" ")[0]
        if not group_id:
            raise ValueError("Something went wrong, group_id was empty.")

        return group_id

    def _set_memory_limits(self, max: str, min_limit: str) -> None:
        self._execute_vsish_command(
            [
                "set",
                f"/sched/groups/{self.group_id}/memAllocationInMB",
                f"max={max}",
                f"minLimit={min_limit}",
            ]
        )

    def _get_memory_scheme(self) -> dict[str, str]:
        """Retrieve and parse the current memory scheme from an ESXi host."""
        mem_sched_allocation = self._execute_vsish_command(
            [
                "get",
                f"/sched/groups/{self.group_id}/memAllocationInMB",
            ]
        )

        data_pattern = re.compile(r"(\w+):\s*(-?\d+)")
        mem_configuration = data_pattern.findall(mem_sched_allocation)
        mem_shed = dict((x, y) for x, y in mem_configuration)
        return mem_shed



from __future__ import annotations

from enum import Enum

from acquire.dynamic.windows.types import OBJECT_DIRECTORY_INFORMATION


class NamedObjectType(Enum):
    ALPC_PORT = "ALPC Port"
    CALLBACK = "Callback"
    DESKTOP = "Desktop"
    DEVICE = "Device"  # NtOpenFile
    DIRECTORY = "Directory"  # NtOpenDirectoryObject
    DRIVER = "Driver"
    ETW_REGISTRATION = "EtwRegistration"
    EVENT = "Event"  # NtOpenEvent
    FILE = "File"  # NtOpenFile
    FILTER_CONNECTION_PORT = "FilterConnectionPort"
    IO_COMPLETION = "IoCompletion"
    IR_TIMER = "IRTimer"
    JOB = "Job"
    KEY = "Key"  # (Zw|Nt)OpenKey
    KEYED_EVENT = "KeyedEvent"
    MUTANT = "Mutant"  # NtOpenMutant
    MUTEX = "Mutex"
    PARTITION = "Partition"
    PROCESS = "Process"
    SECTION = "Section"  # NtOpenSection
    SESSION = "Session"
    SEMAPHORE = "Semaphore"  # (NtOpenSemaphore)
    SYMBOLIC_LINK = "SymbolicLink"  # NtOpenSymbolicLinkObject, NtQuerySymbolicLinkObject
    TIMER = "Timer"  # NtOpenTimer
    THREAD = "Thread"
    TOKEN = "Token"
    TP_WORKER_FACTORY = "TpWorkerFactory"
    TYPE = "Type"
    WAIT_COMPLETION_PACKET = "WaitCompletionPacket"
    WINDOWS_STATION = "WindowStation"

    UNKNOWN = "Unknown"


class NamedObject:
    __slots__ = [
        "root",
        "name",
        "type_name",
    ]

    def __init__(self, root: str, name: str, type_name: NamedObjectType) -> None:
        self.root = root if root.endswith("\\") else f"{root}\\"
        self.name = name
        self.type_name = type_name

    def __repr__(self) -> str:
        return f"{self.root}{self.name}, {self.type_name}"

    @classmethod
    def from_directory_information(
        cls, root_name: str, directory_information: OBJECT_DIRECTORY_INFORMATION
    ) -> NamedObject:
        return cls(
            root=root_name,
            name=directory_information.name,
            type_name=NamedObjectType(directory_information.type_name),
        )



import argparse
import base64
import contextlib
import io
import json
import logging
import multiprocessing
import os
import signal
import sys
import textwrap
from collections import defaultdict, deque
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime, timezone
from pathlib import Path
from queue import Empty as QueueEmptyError
from urllib import request
from urllib.error import HTTPError
from urllib.parse import urljoin

from Crypto.Cipher import AES, PKCS1_OAEP
from Crypto.PublicKey import RSA
from dissect.util.stream import AlignedStream

try:
    from rich.progress import (
        BarColumn,
        DownloadColumn,
        Progress,
        TextColumn,
        TimeRemainingColumn,
        TransferSpeedColumn,
    )

    progress = Progress(
        TextColumn("[bold blue]{task.fields[filename]}", justify="right"),
        BarColumn(bar_width=None),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        DownloadColumn(),
        "•",
        TransferSpeedColumn(),
        "•",
        TimeRemainingColumn(),
        transient=True,
    )
except ImportError:
    progress = None

from acquire.crypt import (
    FILE_MAGIC,
    FILE_VERSION,
    FOOTER_MAGIC,
    HEADER_MAGIC,
    c_acquire,
    key_fingerprint,
)

log = logging.getLogger(__name__)


CHUNK_SIZE = 1024 * 1024 * 4
WORKER_COUNT = max(1, os.cpu_count() - 1)

STATUS_EXIT = 0
STATUS_INFO = 1
STATUS_START = 2
STATUS_UPDATE = 3


class VerifyError(Exception):
    pass


class EncryptedFile(AlignedStream):
    def __init__(self, fh, key_file=None, key_server=None):
        self.fh = fh
        self.key_file = key_file
        self.key_server = key_server

        self._file_header = None
        self._header = None
        self._footer = None
        self.digest = None

        if not key_file and not key_server:
            raise ValueError("Need either key or key server")

        file_header_buf = fh.read(len(c_acquire.file))
        self.file_header = c_acquire.file(file_header_buf)

        header_buf = fh.read(self.file_header.header_size)
        decrypted_header = decrypt_header(
            header_buf, self.file_header.key_digest, key_file=key_file, key_server=key_server
        )
        self.header = c_acquire.header(decrypted_header)

        size = None
        if self.fh.seekable():
            offset = fh.tell()
            fh.seek(-8, io.SEEK_END)
            self.footer = c_acquire.footer(fh)

            fh.seek(-8 - self.footer.length, io.SEEK_END)
            size = fh.tell() - offset
            self.digest = fh.read(self.footer.length)
            fh.seek(offset)

        self.cipher = AES.new(self.header.key, AES.MODE_GCM, nonce=self.header.iv)
        self.cipher.update(file_header_buf + header_buf)

        self.buffers = deque(maxlen=3)

        super().__init__(size)

    def seekable(self):
        return False

    def seek(self, pos, whence=io.SEEK_CUR):
        raise io.UnsupportedOperation("seeking is not allowed")

    def _read(self, offset, length):
        if not self.size:
            result = []

            while length:
                # Triple buffer the reads so that we can catch the last block.
                # Three buffers because the last three reads can be [CHUNK_SIZE, CHUNK_SIZE - n, 0].
                # We only want to check for the footer when we reach EOF in case of streaming or partial data.
                while len(self.buffers) != self.buffers.maxlen and not self.digest:
                    try:
                        chunk = self.fh.read(CHUNK_SIZE)
                    except EOFError:
                        chunk = b""

                    if chunk == b"":
                        last_block = b"".join(self.buffers)

                        footer_size = len(c_acquire.footer)
                        self.footer = c_acquire.footer(last_block[-footer_size:])
                        self.digest = last_block[-footer_size - self.footer.length : -footer_size]

                        self.buffers.clear()
                        self.buffers.append(last_block[: -footer_size - self.footer.length])
                        self.buffers.append(b"")
                        break

                    self.buffers.append(chunk)

                current_chunk = self.buffers.popleft()
                read_size = max(0, length)
                read_chunk = current_chunk[:read_size]
                remainder_chunk = current_chunk[read_size:]
                if remainder_chunk:
                    self.buffers.appendleft(remainder_chunk)

                result.append(read_chunk)
                length -= read_size

            return self.cipher.decrypt(b"".join(result))
        else:
            read_size = max(0, min(length, self.size - offset))
            return self.cipher.decrypt(self.fh.read(read_size))

    def chunks(self, chunk_size=CHUNK_SIZE):
        while True:
            chunk = self.read(chunk_size)
            if not chunk:
                break
            yield chunk

    def verify(self):
        try:
            self.cipher.verify(self.digest)
        except ValueError:
            raise VerifyError("Digest check failed")

    @property
    def file_header(self):
        return self._file_header

    @file_header.setter
    def file_header(self, file_header):
        if file_header.magic != FILE_MAGIC:
            raise ValueError(f"Invalid file magic: {file_header.magic}")

        if file_header.version != FILE_VERSION:
            raise ValueError(f"Unsupported version: {file_header.version}")

        if file_header.header_type != c_acquire.HeaderType.PKCS1_OAEP:
            raise ValueError(f"Unsupported header type: {file_header.header_type}")

        self._file_header = file_header

    @property
    def header(self):
        return self._header

    @header.setter
    def header(self, header):
        if header.magic != HEADER_MAGIC:
            raise ValueError(f"Invalid header magic: {header.magic}")
        self._header = header

    @property
    def footer(self):
        return self._footer

    @footer.setter
    def footer(self, footer):
        if footer.magic != FOOTER_MAGIC:
            raise ValueError(f"Invalid footer magic: {footer}")
        self._footer = footer

    @property
    def timestamp(self):
        return datetime.fromtimestamp(self.file_header.timestamp, timezone.utc)


def decrypt_header(header, fingerprint, key_file=None, key_server=None):
    if not key_file and not key_server:
        raise ValueError("Need either key file or key server")

    if key_file:
        rsa_key = RSA.import_key(key_file.read_text())
        if key_fingerprint(rsa_key.public_key()) != fingerprint:
            raise ValueError("Key doesn't match fingerprint")
        return PKCS1_OAEP.new(rsa_key).decrypt(header)
    else:
        data = json.dumps({"fingerprint": fingerprint.hex(), "header": base64.b64encode(header).decode()}).encode(
            "utf-8"
        )

        url = urljoin(key_server, "/api/v1/decrypt")
        req = request.Request(url, data, headers={"Content-Type": "application/json"})
        try:
            resp = request.urlopen(req)
        except HTTPError as e:
            if e.code == 404:
                raise ValueError("Unknown key fingerprint")
            raise ValueError(f"Failed to decrypt header: {e}")
        result = json.loads(resp.read())

        return base64.b64decode(result["header"])


def worker(task_id, stop_event, status_queue, in_path, out_path, key_file=None, key_server=None, clobber=False):
    try:
        if out_path.exists() and not clobber:
            _info(status_queue, f"Output file already exists: {out_path}")
            return

        _update(status_queue, task_id, visible=True)
        success = False
        with in_path.open("rb") as infh:
            try:
                ef = EncryptedFile(infh, key_file=key_file, key_server=key_server)
            except Exception as e:
                _info(status_queue, f"{in_path} • Error opening encrypted file: {e}")
            _info(
                status_queue,
                f"{in_path} • File: {ef.file_header.magic.decode()} | {ef.file_header.header_type} | {ef.timestamp}",
            )
            _info(status_queue, f"{in_path} • Header: {ef.header.magic.decode()} | {ef.header.cipher_type}")
            _update(status_queue, task_id, total=ef.size)

            with out_path.open("wb") as outfh:
                _info(status_queue, f"{in_path} • Decrypting to {out_path}")
                _start(status_queue, task_id)
                try:
                    for chunk in ef.chunks():
                        if stop_event.is_set():
                            raise ValueError("stopping")
                        outfh.write(chunk)
                        _update(status_queue, task_id, advance=len(chunk))
                    ef.verify()
                    _info(status_queue, f"{in_path} • File verified OK!")
                    success = True
                except ValueError as e:
                    _info(status_queue, f"{in_path} • Error decrypting file: {e}")
                except VerifyError as e:
                    _info(status_queue, f"{in_path} • Verify error: {e}")
                    _info(status_queue, f"{in_path} • ! Output file should NOT be trusted")
                except Exception as e:
                    _info(status_queue, f"{in_path} • Unknown error: {e}")

        if not success:
            failed_path = out_path.with_suffix(f"{out_path.suffix}.failed")
            _info(status_queue, f"{in_path} • ! Renaming to {failed_path}")
            out_path.rename(failed_path)
    finally:
        _update(status_queue, task_id, visible=False)
        _exit(status_queue, task_id)


def _start(queue, task_id):
    queue.put_nowait((STATUS_START, task_id))


def _update(queue, task_id, *args, **kwargs):
    queue.put_nowait((STATUS_UPDATE, (task_id, args, kwargs)))


def _info(queue, msg):
    queue.put_nowait((STATUS_INFO, msg))


def _exit(queue, task_id):
    queue.put_nowait((STATUS_EXIT, task_id))


def setup_logging(logger, verbosity):
    if verbosity == 1:
        level = logging.ERROR
    elif verbosity == 2:
        level = logging.WARNING
    elif verbosity == 3:
        level = logging.INFO
    elif verbosity >= 4:
        level = logging.DEBUG
    else:
        level = logging.CRITICAL

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(logging.Formatter("%(message)s"))
    stream_handler.setLevel(level)
    logger.addHandler(stream_handler)
    logger.setLevel(level)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("files", nargs="+", type=Path, help="paths to encrypted files")
    parser.add_argument("-o", "--output", type=Path, help="optional path to output file")
    parser.add_argument("-c", "--clobber", action="store_true", help="overwrite existing output files")
    parser.add_argument("-ks", "--key-server", default="", help="URI for key server")
    parser.add_argument("-k", "--key-file", type=Path, help="path to PEM private key")
    parser.add_argument("-w", "--workers", type=int, default=WORKER_COUNT, help="maximum amount of workers")
    parser.add_argument("-v", "--verbose", action="count", default=3, help="increase output verbosity")
    args = parser.parse_args()

    setup_logging(log, args.verbose)

    if not progress:
        log.info("`rich` is not installed, progress will not be shown")

    if args.output and args.output.is_file() and len(args.files) > 1:
        parser.exit("--output should be a directory when decrypting multiple files.")

    files = find_enc_files(args.files)

    if args.output:
        resolv_path = args.output.resolve()
        outputs = [resolv_path / path.stem for path in files] if args.output.is_dir() else [resolv_path]
    else:
        # Strip .enc extension
        outputs = [path.with_suffix("") for path in files]

    if len(set(outputs)) != len(outputs):
        show_duplicates(args.output, files)

    if not args.key_file and not args.key_server:
        parser.exit("Need either --key-file or --key-server")

    if args.key_file:
        if not args.key_file.is_file():
            parser.exit(f"{args.key_file} doesn't exist or is not a file")

        key_file = args.key_file
        key_server = None
    else:
        key_file = None
        key_server = args.key_server

    ctx = progress or contextlib.nullcontext()
    with ctx:
        signal_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        with multiprocessing.Manager() as mp, ProcessPoolExecutor(max_workers=args.workers) as executor:
            stop_event = mp.Event()
            status_queue = mp.Queue()
            tasks = []

            for in_path, out_path in zip(files, outputs):
                task_id = (
                    progress.add_task("decrypt", start=False, visible=False, filename=in_path.name)
                    if progress
                    else len(tasks)
                )
                executor.submit(
                    worker,
                    task_id,
                    stop_event,
                    status_queue,
                    in_path,
                    out_path,
                    key_file=key_file,
                    key_server=key_server,
                    clobber=args.clobber,
                )
                tasks.append(task_id)

            signal.signal(signal.SIGINT, signal_handler)

            while tasks:
                try:
                    cmd, args = status_queue.get(timeout=10)

                    if cmd == STATUS_EXIT:
                        tasks.remove(args)
                    elif cmd == STATUS_INFO:
                        if progress:
                            progress.console.log(args)
                        else:
                            log.info(args)
                    elif cmd == STATUS_START:
                        if progress:
                            progress.start_task(args)
                    elif cmd == STATUS_UPDATE:
                        task_id, args, kwargs = args
                        if progress:
                            progress.update(task_id, *args, **kwargs)
                except (QueueEmptyError, KeyboardInterrupt):
                    (progress.console.log if progress else log.info)("Stopping...")
                    stop_event.set()
                    executor.shutdown(wait=True, cancel_futures=True)
                    break


def show_duplicates(output_directory: Path, files: list[Path]) -> None:
    # Gather all files that could cause duplicates in `args.output`.
    input_files = defaultdict(list)
    for input_file in files:
        input_files[input_file.name].append(input_file)

    # Find all duplicates
    duplicate_generator = (file_paths for file_paths in input_files.values() if len(file_paths) > 1)
    duplicates = "\n\n".join(
        textwrap.indent(
            "\n".join(str(file) for file in file_paths),
            prefix="  - ",
        )
        for file_paths in duplicate_generator
    )
    log.warning(
        "Two or more encrypted files have the same name. "
        f"This will skip decrypting the file if it already exists in '{output_directory}'\n"
        f"The files with the same names are:\n"
        f"{duplicates}"
    )


def find_enc_files(files: list[Path]) -> list[Path]:
    encrypted_files = []
    for path in files:
        if path.is_file() and path.suffix == ".enc":
            encrypted_files.append(path)
        elif path.is_dir():
            encrypted_files.extend(path.rglob("*.enc"))
        else:
            log.info(f"File {path!r} does not have the .enc extension. skipping.")
    return encrypted_files


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        pass



from pathlib import Path

import pytest

from acquire.acquire import sort_files


@pytest.mark.parametrize(
    "input, expected_output",
    [
        (
            ["test.log", "test.tar"],
            [Path("test.tar"), Path("test.log")],
        ),
        (
            ["test.json", "test.log", "test.tar"],
            [Path("test.tar"), Path("test.json"), Path("test.log")],
        ),
        (
            ["test.json", "test.log", "test.asdf"],
            [Path("test.asdf"), Path("test.json"), Path("test.log")],
        ),
        (
            ["test3.log", "test2.log", "test1.log"],
            [Path("test1.log"), Path("test2.log"), Path("test3.log")],
        ),
    ],
)
def test_file_sorting(input: list[str], expected_output: list[str]):
    assert sort_files(input) == expected_output
